import numpy as np
import pytest
import tensorflow as tf

from tensorflow_similarity.classification_metrics.f1_score import F1Score
from tensorflow_similarity.classification_metrics.false_positive_rate import FalsePositiveRate
from tensorflow_similarity.classification_metrics.negative_predictive_value import NegativePredictiveValue
from tensorflow_similarity.classification_metrics.precision import Precision
from tensorflow_similarity.classification_metrics.recall import Recall
from tensorflow_similarity.classification_metrics.binary_accuracy import BinaryAccuracy


testdata = [
    # Perfect classification with two distance thresholds
    {'tp': tf.ones((2,)), 'fp': tf.zeros((2,)), 'tn':tf.ones((2,)),
        'fn': tf.zeros((2,)), 'count':2},
    # Inverted classification with two distance thresholds
    {'tp': tf.zeros((2,)), 'fp': tf.ones((2,)), 'tn':tf.zeros((2,)),
        'fn': tf.ones((2,)), 'count':2},
    # Decent classification with two distance thresholds
    {'tp': tf.constant([10.0,20.0]), 'fp': tf.constant([1.0,2.0]),
        'tn': tf.constant([10.0,9.0]), 'fn':tf.constant([15.0,5.0]), 'count':36},
    # Single distance threshold
    {'tp': tf.ones((1,)), 'fp': tf.zeros((1,)), 'tn':tf.ones((1,)),
        'fn': tf.zeros((1,)), 'count':2},
]

test_ids = ['perfect_classification', 'inverted_classification',
        'decent_classification', 'single_threshold']

f1_expected_results = [
    tf.constant([1.0, 1.0]), # Perfect classification with two distance thresholds
    tf.constant([0.0, 0.0]), # Inverted classification with two distance thresholds
    tf.constant([0.555556, 0.851064]), # Decent classification with two distance thresholds
    tf.constant([1.0]), # Single distance threshold
]

@pytest.mark.parametrize(
        "counts, expected",
        zip(testdata, f1_expected_results),
        ids=test_ids)
def test_f1(counts, expected):
    f1 = F1Score()
    results = f1.compute(**counts)
    np.testing.assert_allclose(results.numpy(), expected.numpy(), rtol=1e-5)


precision_expected_results = [
    tf.constant([1.0, 1.0]), # Perfect classification with two distance thresholds
    tf.constant([0.0, 0.0]), # Inverted classification with two distance thresholds
    tf.constant([0.909091, 0.909091]), # Decent classification with two distance thresholds
    tf.constant([1.0]), # Single distance threshold
]

@pytest.mark.parametrize(
        "counts, expected",
        zip(testdata, precision_expected_results),
        ids=test_ids)
def test_precision(counts, expected):
    precision = Precision()
    results = precision.compute(**counts)
    np.testing.assert_allclose(results.numpy(), expected.numpy(), rtol=1e-5)

recall_expected_results = [
    tf.constant([1.0, 1.0]), # Perfect classification with two distance thresholds
    tf.constant([0.0, 0.0]), # Inverted classification with two distance thresholds
    tf.constant([0.4, 0.8]), # Decent classification with two distance thresholds
    tf.constant([1.0]), # Single distance threshold
]

@pytest.mark.parametrize(
        "counts, expected",
        zip(testdata, recall_expected_results),
        ids=test_ids)
def test_recall(counts, expected):
    recall = Recall()
    results = recall.compute(**counts)
    np.testing.assert_allclose(results.numpy(), expected.numpy(), rtol=1e-5)

binary_accuracy_expected_results = [
    tf.constant([0.5, 0.5]), # Perfect classification with two distance thresholds
    tf.constant([0.0, 0.0]), # Inverted classification with two distance thresholds
    tf.constant([0.277778, 0.555556]), # Decent classification with two distance thresholds
    tf.constant([0.5]), # Single distance threshold
]

@pytest.mark.parametrize(
        "counts, expected",
        zip(testdata, binary_accuracy_expected_results),
        ids=test_ids)
def test_binary_accuracy(counts, expected):
    acc = BinaryAccuracy()
    results = acc.compute(**counts)
    np.testing.assert_allclose(results.numpy(), expected.numpy(), rtol=1e-5)

fpr_expected_results = [
    tf.constant([0.0, 0.0]), # Perfect classification with two distance thresholds
    tf.constant([1.0, 1.0]), # Inverted classification with two distance thresholds
    tf.constant([0.090909, 0.181818]), # Decent classification with two distance thresholds
    tf.constant([0.0]), # Single distance threshold
]

@pytest.mark.parametrize(
        "counts, expected",
        zip(testdata, fpr_expected_results),
        ids=test_ids)
def test_false_positive_rate(counts, expected):
    fpr = FalsePositiveRate()
    results = fpr.compute(**counts)
    np.testing.assert_allclose(results.numpy(), expected.numpy(), rtol=1e-5)

npv_expected_results = [
    tf.constant([1.0, 1.0]), # Perfect classification with two distance thresholds
    tf.constant([0.0, 0.0]), # Inverted classification with two distance thresholds
    tf.constant([0.4, 0.642857]), # Decent classification with two distance thresholds
    tf.constant([1.0]), # Single distance threshold
]

@pytest.mark.parametrize(
        "counts, expected",
        zip(testdata, npv_expected_results),
        ids=test_ids)
def test_negative_predicitve_value(counts, expected):
    npv = NegativePredictiveValue()
    results = npv.compute(**counts)
    np.testing.assert_allclose(results.numpy(), expected.numpy(), rtol=1e-5)



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Evaluates search index performance and calibrates it.

## Use

Evaluators are used for two primary purposes:

- Evaluate model performance on a reference index during training and
evaluation phase via the `evaluate_classification()` and `evaluate_retrieval()`
methods. Evaluation on a reference index is
required to be able to assess model performance using
[Classification metrics](../classification_metrics/) and
[Retrieval metrics](../retrieval_metrics/).
Those metrics can't be computed without indexing data and looking up
nearest neighbors.

- Calibrating the model requires evaluating various distance thresholds
to find the maximal distance threshold. Those thresholds either meet,
if possible, the user supplied `thresholds_targets` performance value or
the optimal value with respect to the calibration `classification metric`.
Calibration is required to perform matching
because the optimal distance thresholds will change depending on
the model, dataset and, training. Accordingly those thresholds
need to be evaluated empirically for every use-case.
"""
from .memory_evaluator import MemoryEvaluator  # noqa
from .memory_evaluator import Evaluator  # noqa



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math

import tensorflow as tf

from .retrieval_metric import RetrievalMetric
from tensorflow_similarity.types import FloatTensor, IntTensor, BoolTensor


class BNDCG(RetrievalMetric):
    """Binary normalized discounted cumulative gain.

    This is normalized discounted cumulative gain where the relevancy weights
    are binary, i.e., either a correct match or an incorrect match.

    The NDCG is a score between [0,1] representing the rank weighted results.
    The DCG represents the sum of the correct matches weighted by the log2 of
    the rank and is normalized by the 'ideal DCG'. The IDCG is computed as the
    match_mask, sorted descending, weighted by the log2 of the post sorting rank
    order. This metric takes into account both the correctness of the match and
    the position.

    The normalized DCG is computed as:

    $$
    nDCG_{p} = \frac{DCG_{p}}{IDCG_{p}}
    $$

    The DCG is computed for each query using the match_mask as:

    $$
    DCG_{p} = \sum_{i=1}^{p} \frac{match_mask_{i}}{\log_{2}(i+1)}
    $$

    The IDCG uses the same equation but sorts the match_mask descending
    along axis=-1.

    Additionally, all positive matches with a distance above the threshold are
    set to 0, and the closest K matches are taken.

    Args:
        name: Name associated with the metric object, e.g., precision@5

        canonical_name: The canonical name associated with metric,
        e.g., precision@K

        k: The number of nearest neighbors over which the metric is computed.

        distance_threshold: The max distance below which a nearest neighbor is
        considered a valid match.

        average: {'micro', 'macro'} Determines the type of averaging performed
        on the data.

        * 'micro': Calculates metrics globally over all data.

        * 'macro': Calculates metrics for each label and takes the unweighted
                   mean.
    """

    def __init__(
        self,
        name: str = "ndcg",
        k: int = 5,
        distance_threshold: float = math.inf,
        **kwargs,
    ) -> None:
        if "canonical_name" not in kwargs:
            kwargs["canonical_name"] = "ndcg@k"

        super().__init__(
            name=name, k=k, distance_threshold=distance_threshold, **kwargs
        )

    def compute(
        self,
        *,  # keyword only arguments see PEP-570
        query_labels: IntTensor,
        lookup_distances: FloatTensor,
        match_mask: BoolTensor,
        **kwargs,
    ) -> FloatTensor:
        """Compute the metric

        Computes the binary NDCG. The query labels are only used when the
        averaging is set to "macro".

        Args:
            query_labels: A 1D array of the labels associated with the
            embedding queries.

            lookup_distances: A 2D array where the jth row is the distances
            between the jth query and the set of k neighbors.

            match_mask: A 2D mask where a 1 indicates a match between the
            jth query and the kth neighbor and a 0 indicates a mismatch.

        Returns:
            A rank 0 tensor containing the metric.
        """
        self._check_shape(query_labels, match_mask)

        if tf.shape(lookup_distances)[0] != tf.shape(query_labels)[0]:
            raise ValueError(
                "The number of lookup distance rows must equal the number "
                "of query labels. Number of lookup distance rows is "
                f"{tf.shape(lookup_distances)[0]} but the number of query "
                f"labels is {tf.shape(query_labels)[0]}."
            )

        dist_mask = tf.math.less_equal(
            lookup_distances, self.distance_threshold
        )

        k_slice = tf.math.multiply(
            tf.cast(match_mask, dtype="float"),
            tf.cast(dist_mask, dtype="float"),
        )[:, : self.k]

        rank = tf.range(1, self.k + 1, dtype="float")
        rank_weights = tf.math.divide(tf.math.log1p(rank), tf.math.log(2.0))

        # the numerator is simplier here because we are using binary weights
        dcg = tf.math.reduce_sum(k_slice / rank_weights, axis=1)

        # generate the "ideal ordering".
        ideal_ordering = tf.sort(k_slice, direction="DESCENDING", axis=1)
        idcg = tf.math.reduce_sum(ideal_ordering / rank_weights, axis=1)

        per_example_ndcg = tf.math.divide_no_nan(dcg, idcg)

        if self.average == "micro":
            ndcg = tf.math.reduce_mean(per_example_ndcg)
        elif self.average == "macro":
            per_class_metrics = 0
            class_labels = tf.unique(query_labels)[0]
            for label in class_labels:
                idxs = tf.where(query_labels == label)
                c_slice = tf.gather(per_example_ndcg, indices=idxs)
                per_class_metrics += tf.math.reduce_mean(c_slice)
            ndcg = tf.math.divide(per_class_metrics, len(class_labels))
        else:
            raise ValueError(
                f"{self.average} is not a supported average " "option"
            )

        result: FloatTensor = ndcg
        return result



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple

from .classification_match import ClassificationMatch
from tensorflow_similarity.types import FloatTensor, IntTensor


class MatchNearest(ClassificationMatch):
    """Match metrics for labels at k=1."""

    def __init__(self,
                 name: str = 'nearest',
                 **kwargs) -> None:

        if 'canonical_name' not in kwargs:
            kwargs['canonical_name'] = 'match_nearest'

        super().__init__(name=name, **kwargs)

    def derive_match(self,
                     lookup_labels: IntTensor,
                     lookup_distances: FloatTensor
                     ) -> Tuple[IntTensor, FloatTensor]:
        """Derive a match label and distance from a set of K neighbors.

        For each query, derive a single match label and distance given the
        associated set of lookup labels and distances.

        Args:
            lookup_labels: A 2D array where the jth row is the labels
            associated with the set of k neighbors for the jth query.

            lookup_distances: A 2D array where the jth row is the distances
            between the jth query and the set of k neighbors.

        Returns:
            A Tuple of FloatTensors:
                derived_labels: A FloatTensor of shape
                [len(lookup_labels), 1] where the jth row contains the derived
                label for the jth query.

                derived_distances: A FloatTensor of shape
                [len(lookup_labels), 1] where the jth row contains the distance
                associated with the jth derived label.
        """

        return lookup_labels[:, :1], lookup_distances[:, :1]



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Key Values Stores store the data associated with the embeddings indexed by
the `Indexer()`.

Each key of the store represent a **record** that contains information
about a given embedding.

The main use-case for the store is to retrieve the records associated
with the ids returned by a nearest neigboor search performed with the
[`Search()` module](../search/).

Additionally one might want to inspect the content of the index which is why
`Store()` class may implement an export to
a [Pandas Dataframe](https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html)
via the `to_pandas()` method.
"""

from .store import Store  # noqa
from .memory_store import MemoryStore  # noqa



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from tensorflow_similarity.types import FloatTensor
from .classification_metric import ClassificationMetric


class BinaryAccuracy(ClassificationMetric):
    """Calculates how often the query label matches the derived lookup label.

    Accuracy is technically (TP+TN)/(TP+FP+TN+FN), but here we filter all
    queries above the distance threshold. In the case of binary matching, this
    makes all the TPs and FPs below the distance threshold and all the TNs and
    FNs above the distance threshold.

    As we are only concerned with the matches below the distance threshold, the
    accuracy simplifies to TP/(TP+FP) and is equivalent to the precision with
    respect to the unfiltered queries. However, we also want to consider the
    query coverage at the distance threshold, i.e., the percentage of queries
    that retrun a match, computed as (TP+FP)/(TP+FP+TN+FN). Therefore, we can
    take $ precision \times query_coverage $ to produce a measure that capture
    the precision scaled by the query coverage. This simplifies down to the
    binary accuracy presented here, giving TP/(TP+FP+TN+FN).

    args:
        name: Name associated with a specific metric object, e.g.,
        binary_accuracy@0.1

    Usage with `tf.similarity.models.SimilarityModel()`:

    ```python
    model.calibrate(x=query_examples,
                    y=query_labels,
                    calibration_metric='binary_accuracy')
    ```
    """
    def __init__(self, name: str = 'binary_accuracy') -> None:
        super().__init__(name=name, canonical_name='binary_accuracy')

    def compute(self,
                tp: FloatTensor,
                fp: FloatTensor,
                tn: FloatTensor,
                fn: FloatTensor,
                count: int) -> FloatTensor:
        """Compute the classification metric.

        The `compute()` method supports computing the metric for a set of
        values, where each value represents the counts at a specific distance
        threshold.

        Args:
            tp: A 1D FloatTensor containing the count of True Positives at each
            distance threshold.

            fp: A 1D FloatTensor containing the count of False Positives at
            each distance threshold.

            tn: A 1D FloatTensor containing the count of True Negatives at each
            distance threshold.

            fn: A 1D FloatTensor containing the count of False Negatives at
            each distance threshold.

            count: The total number of queries

        Returns:
            A 1D FloatTensor containing the metric at each distance threshold.
        """
        result: FloatTensor = tp / tf.constant([count], dtype='float')
        return result



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""

![TensorFlow Similarity Overview](../assets/images/tfsim_overview.png)

TensorFlow Similiarity, as visible in the diagram above, offers the following
components to help research, train, evaluate and serve metric models:

- **`SimilarityModel()`**: This class subclasses the `tf.keras.model` class and extends it with additional properties that are useful for metric learning. For example it adds the methods:
  1. `index()`: Enables indexing of the embedding
  2. `lookup()`: Takes samples, calls predict(), and searches for neighbors within the index.
  3. `calibrate()`: Calibrates the model's index search thresholds using a calibration metric and a test dataset.

- **`MetricLoss()`**:  This virtual class, that extends the `tf.keras.Loss` class, is the base class from which Metric losses are derived. This sub-classing ensures proper error checking; that is, it ensures the user is using a loss metric to train the models, performs better static analysis, and enforces additional constraints such as having a distance function that is supported by the index. Additionally, Metric losses make use of the fully tested and highly optimized pairwise distances functions provided by TensorFlow Similarity that are available under the `Distances.*` classes.

- **`Samplers()`**: Samplers are meant to ensure that each batch has at least n (with n >=2) examples of each class, as losses such as TripletLoss can’t work properly if this condition is not met. TensorFlow Similarity offers an in-memory sampler for small dataset and a `tf.data.TFRecordDataset` for large scales one.

- **`Indexer()`**: The Indexer and its sub-components are meant to index known embeddings alongside their metadata. The embedding metadata is stored within `Table()`, while the `Matcher()` is used to perform [fast approximate neighbor searches](https://en.wikipedia.org/wiki/Nearest_neighbor_search) that are meant to quickly retrieve the indexed elements that are the closest to the embeddings supplied in the `lookup()` and `single_lookup()` function.

The default `Index()` sub-compoments run in-memory and are optimized to be used in interactive settings such as Jupyter notebooks, Colab, and metric computation during training (e.g using the `EvalCallback()` provided). Index are serialized as part of `model.save()` so you can reload them via `model.index_load()` for serving purpose or further training / evaluation.

The default implementation can scale up to medium deployment (1M-10M+ points) easily, provided the computers have enough memory. For very large scale deployments you will need to sublcass the compoments to match your own architetctue. See FIXME colab to see how to deploy TensorFlow Similarity in production.
"""

# The file is used to list the public API for document generation purpose.
from tensorflow_similarity import architectures  # noqa
from tensorflow_similarity import evaluators  # noqa
from tensorflow_similarity import layers  # noqa
from tensorflow_similarity import losses  # noqa
from tensorflow_similarity import search  # noqa
from tensorflow_similarity import matchers  # noqa
from tensorflow_similarity import models  # noqa
from tensorflow_similarity import samplers  # noqa
from tensorflow_similarity import stores  # noqa
from tensorflow_similarity import visualization  # noqa
# from tensorflow_similarity import algebra  # not public
from tensorflow_similarity import callbacks  # noqa
from tensorflow_similarity import distances  # noqa
from tensorflow_similarity import indexer  # noqa
from tensorflow_similarity import classification_metrics  # noqa
from tensorflow_similarity import retrieval_metrics  # noqa
from tensorflow_similarity import training_metrics  # noqa
# from tensorflow_similarity import types  # not public
from tensorflow_similarity import utils  # noqa



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import io
from typing import Any, List, Mapping, Sequence, Optional

from distinctipy import distinctipy
import numpy as np
import PIL
import umap
from bokeh.plotting import ColumnDataSource, figure, show, output_notebook
from tqdm.auto import tqdm

from tensorflow_similarity.types import FloatTensor, Tensor


def tensor2images(tensor: Tensor, size: Optional[int] = 64) -> List[str]:
    """Convert tensor images back to in memory images
    encoded in base 64.

    Args:
        tensor: 4D tensor that represent an image list.
        size: Image size to output in pixels. Defaults to 64.

    Returns:
        list of images encoded as base64 strings
    """

    # casting as iterating over a Tensor is slow.
    data = np.array(tensor)

    # if image provided are scaled between [0,1] then rescale
    if np.max(data) <= 1:
        data = data * 255

    # cast as int so PIL accepts its
    data = np.uint8(data)

    imgs_b64 = []
    for a in tqdm(data, desc="generating diplayabe images"):
        # if single channel, treat it as black and white
        if a.shape[-1] == 1:
            a = np.reshape(a, (a.shape[0], a.shape[1]))
            img = PIL.Image.fromarray(a, 'L')
        else:
            img = PIL.Image.fromarray(a)

        img_resized = img.resize((size, size))
        buffer = io.BytesIO()
        img_resized.save(buffer, format='JPEG')
        img_bytes = buffer.getvalue()
        img64 = 'data:image/png;base64,%s' % str(
            base64.b64encode(img_bytes))[2:-1]
        imgs_b64.append(img64)

    return imgs_b64


def projector(embeddings: FloatTensor,
              labels: Optional[Sequence[Any]] = None,
              class_mapping: Optional[Sequence[int]] = None,
              images: Optional[Tensor] = None,
              image_size: int = 64,
              tooltips_info: Optional[Mapping[str, Sequence[str]]] = None,
              pt_size: int = 3,
              colorize: bool = True,
              pastel_factor: float = 0.1,
              plot_size: int = 600,
              active_drag: str = 'box_zoom',
              densmap: bool = True):
    """Visualize the embeddings in 2D or 3D using UMAP projection

    Args:
        embeddings: The embeddings outputed by the model that
        are to be visualized

        labels: Labels associated with the embeddings. If not supplied treat
        each example as its own classes.

        class_mapping: Dictionary or list that maps the class numerical ids
        to their name.

        images: Images to display in tooltip on hover. Usually x_test tensor.

        image_size: size of the images displayed in the tool tip.
        Defaults to 64.

        pt_size: Size of the points displayed on the visualization.
        Defaults to 3.

        tooltips_info: Dictionary of information to display in the tooltips.

        colorize: Colorize the clusters. Defaults to true.

        pastel_factor: Modify the color palette to be more pastel.

        densmap: Use UMAP dense mapper which provides better density
        estimation but is a little slower. Defaults to True.
    """

    print("perfoming projection using UMAP")
    reducer = umap.UMAP(densmap=densmap)
    # FIXME: 2d vs 3d
    cords = reducer.fit_transform(embeddings)

    # sample id
    _idxs = [i for i in range(len(embeddings))]

    # labels?
    if labels is not None:
        # if labels are already names just use them.
        if isinstance(labels[0], str):
            _labels = labels
        else:
            _labels = [int(i) for i in labels]
    else:
        # treat each examples as its own class
        _labels = _idxs

    # class name mapping?
    if class_mapping:
        _labels_txt = [class_mapping[i] for i in _labels]
    else:
        _labels_txt = [str(i) for i in _labels]

    class_list = sorted(set(_labels_txt))
    num_classes = len(class_list)

    # generate data
    data = dict(
        id=_idxs,
        x=[i[0] for i in cords],
        y=[i[1] for i in cords],
        labels=_labels,
        labels_txt=_labels_txt,
    )

    # colors if needed
    if labels is not None and colorize:
        # generate colors
        colors = {}
        for idx, c in enumerate(
                distinctipy.get_colors(num_classes,
                                       pastel_factor=pastel_factor)):
            # this is needed as labels can be strings or int or else
            cls_id = class_list[idx]
            colors[cls_id] = distinctipy.get_hex(c)

        # map point to their color
        _colors = [colors[i] for i in _labels_txt]
        data['colors'] = _colors
    else:
        _colors = []

    # building custom tooltips
    tooltips = '<div style="border:1px solid #ABABAB">'

    if images is not None:
        imgs = tensor2images(images, image_size)
        data['imgs'] = imgs
        # have to write custom tooltip html.
        tooltips += '<center><img src="@imgs"/></center>'  # noqa

    # adding user info
    if tooltips_info:
        for k, v in tooltips_info.items():
            data[k] = v
            tooltips += "%s:@%s <br>" % (k, k)

    tooltips += 'Class:@labels_txt <br>ID:@id </div>'

    # to bokeh format
    source = ColumnDataSource(data=data)
    output_notebook()
    fig = figure(tooltips=tooltips,
                 plot_width=plot_size,
                 plot_height=plot_size,
                 active_drag=active_drag,
                 active_scroll="wheel_zoom")

    # remove grid and axis
    fig.xaxis.visible = False
    fig.yaxis.visible = False
    fig.xgrid.visible = False
    fig.ygrid.visible = False

    # draw points
    if len(_colors):
        fig.circle('x', 'y', size=pt_size, color='colors', source=source)
    else:
        fig.circle('x', 'y', size=pt_size, source=source)

    # render
    output_notebook()
    show(fig, notebook_handle=True)



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import List, Optional, Sequence, Tuple

import numpy as np
import pandas as pd

from tensorflow_similarity.types import FloatTensor, PandasDataFrame, Tensor
from .store import Store


class MemoryStore(Store):
    """Efficient in-memory dataset store"""

    def __init__(self) -> None:
        # We are using a native python array in memory for its row speed.
        # Serialization / export relies on Arrow.
        self.labels: List[Optional[int]] = []
        self.embeddings: List[FloatTensor] = []
        self.data: List[Optional[Tensor]] = []
        self.num_items: int = 0
        pass

    def add(self,
            embedding: FloatTensor,
            label: Optional[int] = None,
            data: Optional[Tensor] = None) -> int:
        """Add an Embedding record to the key value store.

        Args:
            embedding: Embedding predicted by the model.

            label: Class numerical id. Defaults to None.

            data: Data associated with the embedding. Defaults to None.

        Returns:
            Associated record id.
        """
        idx = self.num_items
        self.labels.append(label)
        self.embeddings.append(embedding)
        self.data.append(data)
        self.num_items += 1
        return idx

    def batch_add(self,
                  embeddings: Sequence[FloatTensor],
                  labels: Optional[Sequence[int]] = None,
                  data: Optional[Sequence[Tensor]] = None) -> List[int]:
        """Add a set of embedding records to the key value store.

        Args:
            embeddings: Embeddings predicted by the model.

            labels: Class numerical ids. Defaults to None.

            data: Data associated with the embeddings. Defaults to None.

        See:
            add() for what a record contains.

        Returns:
            List of associated record id.
        """
        idxs: List[int] = []
        for idx, embedding in enumerate(embeddings):
            label = None if labels is None else labels[idx]
            rec_data = None if data is None else data[idx]
            idxs.append(self.add(embedding, label, rec_data))
        return idxs

    def get(self, idx: int) -> Tuple[FloatTensor,
                                     Optional[int],
                                     Optional[Tensor]]:
        """Get an embedding record from the key value store.

        Args:
            idx: Id of the record to fetch.

        Returns:
            record associated with the requested id.
        """

        return self.embeddings[idx], self.labels[idx], self.data[idx]

    def batch_get(self,
                  idxs: Sequence[int]) -> Tuple[List[FloatTensor],
                                                List[Optional[int]],
                                                List[Optional[Tensor]]]:
        """Get embedding records from the key value store.

        Args:
            idxs: ids of the records to fetch.

        Returns:
            List of records associated with the requested ids.
        """
        embeddings = []
        labels = []
        data = []
        for idx in idxs:
            e, l, d = self.get(idx)
            embeddings.append(e)
            labels.append(l)
            data.append(d)
        return embeddings, labels, data

    def size(self) -> int:
        "Number of record in the key value store."
        return self.num_items

    def save(self, path: str, compression: bool = True) -> None:
        """Serializes index on disk.

        Args:
            path: where to store the data.
            compression: Compress index data. Defaults to True.
        """
        fname = self._make_fname(path)
        if compression:
            np.savez_compressed(fname,
                                embeddings=self.embeddings,
                                labels=self.labels,
                                data=self.data)
        else:
            np.savez(fname,
                     embeddings=self.embeddings,
                     labels=self.labels,
                     data=self.data)

    def load(self, path: str) -> int:
        """load index on disk

        Args:
            path: which directory to use to store the index data.

        Returns:
           Number of records reloaded.
        """

        fname = self._make_fname(path, check_file_exit=True)
        data = np.load(fname, allow_pickle=True)
        self.embeddings = list(data['embeddings'])
        self.labels = list(data['labels'])
        self.data = list(data['data'])
        self.num_items = len(self.embeddings)
        print("loaded %d records from %s" % (self.size(), path))
        return self.size()

    def _make_fname(self, path: str, check_file_exit: bool = False) -> str:
        p = Path(path)
        if not p.exists():
            raise ValueError("Index path doesn't exist")
        fname = p / 'index.npz'

        # only for loading
        if check_file_exit and not fname.exists():
            raise ValueError("Index file not found")
        return str(fname)

    def to_data_frame(self, num_records: int = 0) -> PandasDataFrame:
        """Export data as a Pandas dataframe.

        Args:
            num_records: Number of records to export to the dataframe.
            Defaults to 0 (unlimited).

        Returns:
            pd.DataFrame: a pandas dataframe.
        """

        if not num_records:
            num_records = self.num_items

        data = {
            "embeddings": self.embeddings[:num_records],
            "data": self.data[:num_records],
            "lables": self.labels[:num_records]
        }

        # forcing type from Any to PandasFrame
        df: PandasDataFrame = pd.DataFrame.from_dict(data)
        return df



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from copy import copy
import math
from typing import DefaultDict, Dict, List, MutableMapping, Sequence, Union

import numpy as np
import tensorflow as tf
from tqdm.auto import tqdm

from .evaluator import Evaluator
from tensorflow_similarity.classification_metrics import ClassificationMetric
from tensorflow_similarity.matchers import ClassificationMatch
from tensorflow_similarity.matchers import make_classification_matcher
from tensorflow_similarity.retrieval_metrics import RetrievalMetric
from tensorflow_similarity.retrieval_metrics.utils import compute_match_mask
from tensorflow_similarity.types import (
        Lookup, CalibrationResults, IntTensor, FloatTensor)
from tensorflow_similarity.utils import unpack_lookup_distances
from tensorflow_similarity.utils import unpack_lookup_labels


class MemoryEvaluator(Evaluator):
    """In memory index performance evaluation and classification."""

    def evaluate_retrieval(
            self,
            target_labels: Sequence[int],
            lookups: Sequence[Sequence[Lookup]],
            retrieval_metrics: Sequence[RetrievalMetric],
            distance_rounding: int = 8) -> Dict[str, np.ndarray]:
        """Evaluates lookup performances against a supplied set of metrics

        Args:
            target_labels: Sequence of the expected labels to match.

            lookups: Sequence of lookup results as produced by the
            `Index().batch_lookup()` method.

            retrieval_metrics: Sequence of `RetrievalMetric()` to evaluate
            lookup matches against.

            distance_rounding: How many digit to consider to decide if
            the distance changed. Defaults to 8.

        Returns:
            Dictionary of metric results where keys are the metric names and
            values are the metrics values.
        """
        # data preparation: flatten and rounding
        # lookups will be shape(num_queries, num_neighbors)
        # distances will be len(num_queries x num_neighbors)
        nn_labels = unpack_lookup_labels(lookups)
        distances = unpack_lookup_distances(lookups, distance_rounding)

        lookup_set_size = tf.shape(nn_labels)[1]
        for m in retrieval_metrics:
            if lookup_set_size < m.k:
                raise ValueError(
                        f'Each query example returned {lookup_set_size} '
                        f'neighbors, but retrieval metric {m.name} '
                        f'requires the K >= {m.k}.')

        # ensure the target labels are an int32 tensor
        query_labels: IntTensor = tf.cast(
                tf.convert_to_tensor(target_labels),
                dtype='int32'
        )
        match_mask = compute_match_mask(query_labels, nn_labels)

        # compute metrics
        evaluation = {}
        for m in retrieval_metrics:
            res = m.compute(
                query_labels=query_labels,
                lookup_labels=nn_labels,
                lookup_distances=distances,
                match_mask=match_mask
            )
            evaluation[m.name] = res.numpy()

        return evaluation

    def evaluate_classification(
        self,
        query_labels: IntTensor,
        lookup_labels: IntTensor,
        lookup_distances: FloatTensor,
        distance_thresholds: FloatTensor,
        metrics: Sequence[ClassificationMetric],
        matcher: Union[str, ClassificationMatch],
        distance_rounding: int = 8,
        verbose: int = 1
    ) -> Dict[str, np.ndarray]:
        """Evaluate the classification performance.

        Compute the classification metrics given a set of queries, lookups, and
        distance thresholds.

        Args:
            query_labels: Sequence of expected labels for the lookups.

            lookup_labels: A 2D tensor where the jth row is the labels
            associated with the set of k neighbors for the jth query.

            lookup_distances: A 2D tensor where the jth row is the distances
            between the jth query and the set of k neighbors.

            distance_thresholds: A 1D tensor denoting the distances points at
            which we compute the metrics.

            metrics: The set of classification metrics.

            matcher: {'match_nearest', 'match_majority_vote'} or
            ClassificationMatch object. Defines the classification matching,
            e.g., match_nearest will count a True Positive if the query_label
            is equal to the label of the nearest neighbor and the distance is
            less than or equal to the distance threshold.

            distance_rounding: How many digit to consider to
            decide if the distance changed. Defaults to 8.

            verbose: Be verbose. Defaults to 1.

        Returns:
            A Mapping from metric name to the list of values computed for each
            distance threshold.
        """
        matcher = make_classification_matcher(matcher)
        matcher.compile(distance_thresholds=distance_thresholds)

        # compute the tp, fp, tn, fn counts
        matcher.compute_count(
                query_labels=query_labels,
                lookup_labels=lookup_labels,
                lookup_distances=lookup_distances)

        # evaluating performance as distance value increase
        if verbose:
            pb = tqdm(total=len(metrics), desc='Evaluating')

        # evaluating performance as distance value increase
        results: Dict[str, np.ndarray] = (
                {'distance': distance_thresholds.numpy()})
        for m in metrics:
            res = m.compute(
                    tp=matcher.tp,
                    fp=matcher.fp,
                    tn=matcher.tn,
                    fn=matcher.fn,
                    count=matcher.count)
            results[m.name] = res.numpy()

            if verbose:
                pb.update()

        if verbose:
            pb.close()

        return results

    def calibrate(
        self,
        target_labels: Sequence[int],
        lookups: Sequence[Sequence[Lookup]],
        thresholds_targets: MutableMapping[str, float],
        calibration_metric: ClassificationMetric,
        matcher: Union[str, ClassificationMatch],
        extra_metrics: Sequence[ClassificationMetric] = [],
        distance_rounding: int = 8,
        metric_rounding: int = 6,
        verbose: int = 1
    ) -> CalibrationResults:
        """Computes the distances thresholds that the classification must match to
        meet a fixed target.

        Args:
            target_labels: Sequence of expected labels for the lookups.

            lookup: Sequence of lookup results as produced by the
            `Index.batch_lookup()` method.

            thresholds_targets: classification metrics thresholds that are
            targeted. The function will find the closed distance value.

            calibration_metric: Classification metric used for calibration.

            matcher: {'match_nearest', 'match_majority_vote'} or
            ClassificationMatch object. Defines the classification matching,
            e.g., match_nearest will count a True Positive if the query_label
            is equal to the label of the nearest neighbor and the distance is
            less than or equal to the distance threshold.

            extra_metrics: Additional classification metrics that should be
            computed and reported as part of the calibration. Defaults to [].

            distance_rounding: How many digit to consider to
            decide if the distance changed. Defaults to 8.

            metric_rounding: How many digit to consider to decide if
            the metric changed. Defaults to 6.

            verbose: Be verbose. Defaults to 1.
        Returns:
            CalibrationResults containing the thresholds and cutpoints Dicts.
        """

        # making a single list of metrics
        # Need expl covariance problem
        combined_metrics = list(extra_metrics)
        combined_metrics.append(calibration_metric)

        # data preparation: flatten and rounding
        # lookups will be shape(num_queries, num_neighbors)
        # distances will be len(num_queries x num_neighbors)
        lookup_distances = unpack_lookup_distances(lookups, distance_rounding)
        lookup_labels = unpack_lookup_labels(lookups)
        # ensure the target labels are an int32 tensor
        query_labels: IntTensor = tf.cast(
                tf.convert_to_tensor(target_labels),
                dtype='int32'
        )

        # the unique set of distance values sorted ascending
        distance_thresholds = tf.sort(
                tf.unique(
                    tf.reshape(lookup_distances, (-1))
                )[0]  # we only use the y output from tf.unique
        )

        results = self.evaluate_classification(
            query_labels=query_labels,
            lookup_labels=lookup_labels,
            lookup_distances=lookup_distances,
            distance_thresholds=distance_thresholds,
            metrics=combined_metrics,
            matcher=matcher,
            distance_rounding=distance_rounding,
            verbose=verbose)

        # pack results into one dict per dist_threshold
        # TODO(ovallis): we can likey refactor the rest of this method to
        # remove the need for this unpacking, but keeping it the same for now
        # to ensure the code is working as expected.
        evaluations = []
        for i, dist in enumerate(distance_thresholds.numpy()):
            ev = {'distance': results['distance'][i]}
            for m in combined_metrics:
                ev[m.name] = results[m.name][i]

            evaluations.append(ev)

        # find the thresholds by going from right to left

        # distances are rounded because of numerical instablity
        # copy threshold targets as we are going to delete them and don't want
        # to alter users supplied data
        thresholds_targets = copy(thresholds_targets)

        # which direction metric improvement is?
        # !loop is right to left so max is decreasing and min is increasing
        if calibration_metric.direction == 'max':
            # we want the lowest value at the largest distance possible
            cmp = self._is_lower
            prev_value = math.inf  # python 3.x only
        else:
            # we want the highest value at the largest distance possible
            cmp = self._is_higher
            prev_value = 0

        # we need a collection of list to apply vectorize operations and make
        # the analysis / viz of the classification data signifcantly easier
        thresholds: DefaultDict[str, List[Union[float, int]]] = defaultdict(list)  # noqa
        cutpoints: DefaultDict[str, Dict[str, Union[str, float, int]]] = defaultdict(dict)  # noqa
        num_distances = len(distance_thresholds)

        if verbose:
            pb = tqdm(total=num_distances, desc='computing thresholds')

        # looping from right to left as we want the max distance for a given
        # metric value
        for ridx in range(num_distances):
            idx = num_distances - ridx - 1  # reversed

            # Rounding the classification metric to create bins
            curr_eval: Dict[str, float] = evaluations[idx]
            classification_value = curr_eval[calibration_metric.name]
            curr_value = round(classification_value, metric_rounding)

            # ? if bug use this line check that the values evolve correclty.
            # print(curr_value, prev_value, cmp(curr_value, prev_value))

            if cmp(curr_value, prev_value):

                # add a new distance threshold
                # cast numpy float32 to python float to make it json
                # serializable
                thresholds['value'].append(float(curr_value))

                # ! the correct distance is already in the eval data
                # record the value for all the metrics requested by the user
                for key, val in curr_eval.items():
                    # cast numpy float32 to python float to make it json
                    # serializable
                    thresholds[key].append(float(val))

                # update current threshold value
                prev_value = curr_value

                # check if the current value meet or exceed threshold target
                to_delete = []  # can't delete in an iteration loop
                for name, value in thresholds_targets.items():
                    if cmp(curr_value, value, equal=True):
                        cutpoints[name] = {'name': name}  # useful for display
                        for k in thresholds.keys():
                            cutpoints[name][k] = thresholds[k][-1]
                        to_delete.append(name)

                # removing found targets to avoid finding lower value
                # recall we go from right to left in the evaluation
                for name in to_delete:
                    del thresholds_targets[name]

            if verbose:
                pb.update()

        if verbose:
            pb.close()

        # find the optimal cutpoint
        if calibration_metric.direction == 'min':
            best_idx = tf.math.argmin(thresholds[calibration_metric.name])
        else:
            best_idx = tf.math.argmax(thresholds[calibration_metric.name])

        # record its value
        cutpoints['optimal'] = {'name': 'optimal'}  # useful for display
        for k in thresholds.keys():
            cutpoints['optimal'][k] = thresholds[k][best_idx]

        # reverse the threshold so they go from left to right as user expect
        for k in thresholds.keys():  # this syntax is need for mypy ...
            thresholds[k].reverse()

        return CalibrationResults(cutpoints=cutpoints, thresholds=thresholds)

    def _is_lower(self, curr, prev, equal=False):
        if equal:
            return curr <= prev
        return curr < prev

    def _is_higher(self, curr, prev, equal=False):
        if equal:
            return curr >= prev
        return curr > prev



import numpy as np
import tensorflow as tf

from tensorflow_similarity.matchers import MatchMajorityVote


def test_predict():
    mn = MatchMajorityVote()

    lookup_labels = tf.constant(
        [
            [10, 12, 10, 12, 10],
            # Ties should take the closer label.
            [20, 13, 13, 20, 30],
        ]
    )
    lookup_distances = tf.constant(
        [[1.0, 1.1, 1.2, 1.3, 1.4], [2.0, 2.1, 2.2, 2.3, 2.4]]
    )

    d_labels, d_dist = mn.derive_match(lookup_labels, lookup_distances)

    np.testing.assert_array_equal(d_labels.numpy(), np.array([[10], [20]]))
    np.testing.assert_allclose(d_dist.numpy(), np.array([[1.2], [2.2]]))



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple

import tensorflow as tf

from abc import abstractmethod
from abc import ABC
import math

from tensorflow_similarity.types import FloatTensor, IntTensor, BoolTensor


class ClassificationMatch(ABC):
    """Abstract base class for defining the classification matching strategy.

    Attributes:
        name: Name associated with the match object, e.g., match_nearest

        canonical_name: The canonical name associated with match strategy,
        e.g., match_nearest

        distance_thresholds: The max distance below which a nearest neighbor is
        considered a valid match. Defaults to None and must be set using
        `ClassificationMatch.compile()`.
    """

    def __init__(self,
                 name: str = '',
                 canonical_name: str = '') -> None:
        self.name = name
        self.canonical_name = canonical_name
        self.distance_thresholds = None

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return "%s:%s" % (self.canonical_name, self.name)

    def get_config(self):
        return {
            "name": str(self.name),
            "canonical_name": str(self.canonical_name),
            "distance_thresholds": self.distance_thresholds,
        }

    def compile(self,
                distance_thresholds: FloatTensor = tf.constant([math.inf])):
        """Configures the distance thresholds used during matching."""
        self.distance_thresholds = tf.sort(
                tf.cast(distance_thresholds, dtype='float32')
        )

    @abstractmethod
    def derive_match(self,
                     lookup_labels: IntTensor,
                     lookup_distances: FloatTensor
                     ) -> Tuple[IntTensor, FloatTensor]:
        """Derive a match label and distance from a set of K neighbors.

        For each query, derive a single match label and distance given the
        associated set of lookup labels and distances.

        Args:
            lookup_labels: A 2D array where the jth row is the labels
            associated with the set of k neighbors for the jth query.

            lookup_distances: A 2D array where the jth row is the distances
            between the jth query and the set of k neighbors.

        Returns:
            A Tuple of FloatTensors:
                derived_labels: A FloatTensor of shape
                [len(lookup_labels), 1] where the jth row contains the derived
                label for the jth query.

                derived_distances: A FloatTensor of shape
                [len(lookup_labels), 1] where the jth row contains the distance
                associated with the jth derived label.
        """

    def _compute_match_indicators(self,
                                  query_labels: IntTensor,
                                  lookup_labels: IntTensor,
                                  lookup_distances: FloatTensor
                                  ) -> Tuple[BoolTensor, BoolTensor]:
        """Compute the match indicator tensor.

        Compute the match indicator tensor given a set of query labels and a
        set of associated lookup labels and distances.

        The method first calls `derive_match()` on the lookup labels and
        distances and then compares the query labels (y_true) against the
        derived labels (y_pred) and checks if the derived distance is <= to the
        distance thresholds.

        Args:
            query_labels: A 1D array of the labels associated with the queries.

            lookup_labels: A 2D array where the jth row is the labels
            associated with the set of k neighbors for the jth query.

            lookup_distances: A 2D array where the jth row is the distances
            between the jth query and the set of k neighbors.

        Returns:
            A Tuple of BoolTensors:
                match_mask: A BoolTensor of shape [len(query_labels), 1]. True
                if the match label == query label, False otherwise.

                distance_mask: A BoolTensor of shape [len(query_labels),
                len(distance_thresholds)]. True if the distance of the jth
                match <= the kth distance threshold.
        """
        if tf.rank(query_labels) == 1:
            query_labels = tf.expand_dims(query_labels, axis=-1)

        ClassificationMatch._check_shape(
                query_labels,
                lookup_labels,
                lookup_distances
        )

        d_labels, d_dist = self.derive_match(lookup_labels, lookup_distances)

        # Safety check to ensure that self.derive_match returns the right
        # shapes.
        if tf.rank(d_labels) == 1:
            d_labels = tf.expand_dims(d_labels, axis=-1)
        if tf.rank(d_dist) == 1:
            d_dist = tf.expand_dims(d_dist, axis=-1)

        # A 1D BoolTensor [len(query_labels), 1]
        match_mask = tf.math.equal(d_labels, query_labels)

        # A 2D BoolTensor [len(lookup_distance), len(self.distance_thresholds)]
        distance_mask = tf.math.less_equal(d_dist, self.distance_thresholds)

        return match_mask, distance_mask

    def compute_count(self,
                      query_labels: IntTensor,
                      lookup_labels: IntTensor,
                      lookup_distances: FloatTensor) -> None:
        """Computes the match counts at each of the distance thresholds.

        This method computes the following at each distance threshold.

        * True Positive: The query label matches the derived lookup label and
        the derived lookup distance is <= the current distance threshold.

        * False Positive: The query label does not match the derived lookup
        label but the derived lookup distance is <= the current distance
        threshold.

        * False Negative: The query label matches the derived lookup label but
        the derived lookup distance is > the current distance threshold.

        * True Negative: The query label does not match the derived lookup
        label and the derived lookup distance is > the current distance
        threshold.

        Note: compile must be called before calling match.

        Args:
            query_labels: A 1D array of the labels associated with the queries.

            lookup_labels: A 2D array where the jth row is the labels
            associated with the set of k neighbors for the jth query.

            lookup_distances: A 2D array where the jth row is the distances
            between the jth query and the set of k neighbors.
        """
        match_mask, distance_mask = self._compute_match_indicators(
                query_labels=query_labels,
                lookup_labels=lookup_labels,
                lookup_distances=lookup_distances
        )

        self._compute_count(match_mask, distance_mask)

    def _compute_count(self,
                       label_match: BoolTensor,
                       dist_mask: BoolTensor) -> None:
        _tp = tf.math.logical_and(label_match, dist_mask)
        _tp = tf.math.count_nonzero(_tp, axis=0)
        self._tp: FloatTensor = tf.cast(_tp, dtype='float')

        _fn = tf.math.logical_and(
                label_match,
                tf.math.logical_not(dist_mask)
        )
        _fn = tf.math.count_nonzero(_fn, axis=0)
        self._fn: FloatTensor = tf.cast(_fn, dtype='float')

        _fp = tf.math.logical_and(
                tf.math.logical_not(label_match),
                dist_mask
        )
        _fp = tf.math.count_nonzero(_fp, axis=0)
        self._fp: FloatTensor = tf.cast(_fp, dtype='float')

        _tn = tf.math.logical_and(
                tf.math.logical_not(label_match),
                tf.math.logical_not(dist_mask)
        )
        _tn = tf.math.count_nonzero(_tn, axis=0)
        self._tn: FloatTensor = tf.cast(_tn, dtype='float')

        self._count = len(label_match)

    @property
    def tp(self) -> FloatTensor:
        """The count of True positive matches.

        A True positive match is when the query label == the label generated by
        a matcher and the distance to the query is less than the distance
        threshold.

        Raises:
            AttributeError: Matcher.compute_count() must be called before
            accessing counts.
        """
        try:
            return self._tp
        except AttributeError as attribute_error:
            raise AttributeError(
                    'Matcher.compute_count() must be called before '
                    'accessing the counts.') from attribute_error

    @property
    def fp(self) -> FloatTensor:
        """The count of False positive matches.

        A False positive match is when the query label != the label generated
        by a matcher and the distance to the query is less than the distance
        threshold.

        Raises:
            AttributeError: Matcher.compute_count() must be called before
            accessing counts.
        """
        try:
            return self._fp
        except AttributeError as attribute_error:
            raise AttributeError(
                    'Matcher.compute_count() must be called before '
                    'accessing the counts.') from attribute_error

    @property
    def tn(self) -> FloatTensor:
        """The count of True negatives matches.

        A True negative match is when the query label != the label generated
        by a matcher and the distance to the query is greater than the distance
        threshold.

        Raises:
            AttributeError: Matcher.compute_count() must be called before
            accessing counts.
        """
        try:
            return self._tn
        except AttributeError as attribute_error:
            raise AttributeError(
                    'Matcher.compute_count() must be called before '
                    'accessing the counts.') from attribute_error

    @property
    def fn(self) -> FloatTensor:
        """The count of False negatives matches.

        A False negative match is when the query label == the label generated
        by a matcher and the distance to the query is greater than the distance
        threshold.

        Raises:
            AttributeError: Matcher.compute_count() must be called before
            accessing counts.
        """
        try:
            return self._fn
        except AttributeError as attribute_error:
            raise AttributeError(
                    'Matcher.compute_count() must be called before '
                    'accessing the counts.') from attribute_error

    @property
    def count(self) -> int:
        """The total number of queries.

        Raises:
            AttributeError: Matcher.compute_count() must be called before
            accessing counts.
        """
        try:
            return self._count
        except AttributeError as attribute_error:
            raise AttributeError(
                    'Matcher.compute_count() must be called before '
                    'accessing the counts.') from attribute_error

    @staticmethod
    def _check_shape(query_labels, lookup_labels, lookup_distances) -> bool:
        if tf.rank(lookup_labels) != 2:
            raise ValueError('lookup_labels must be a 2D tensor of '
                             'shape [len(query_labels), K].')
        if tf.rank(lookup_distances) != 2:
            raise ValueError('lookup_distances must be a 2D tensor of '
                             'shape [len(query_labels), K].')

        q_shape = tf.shape(query_labels)
        ll_shape = tf.shape(lookup_labels)
        ld_shape = tf.shape(lookup_distances)

        if q_shape[0] != ll_shape[0]:
            raise ValueError('Number of query labels must match the number of '
                             'lookup_label sets.')

        if ll_shape[0] != ld_shape[0] or ll_shape[1] != ld_shape[1]:
            raise ValueError('Number of number of lookup labels must match '
                             'the number of lookup distances.')

        return True



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Specialized Similarity `keras.layers`"""


from typing import Dict
import tensorflow as tf
from tensorflow.keras.layers import Layer
from tensorflow.keras import layers
from .types import FloatTensor

@tf.keras.utils.register_keras_serializable(package="Similarity")
class MetricEmbedding(Layer):
    def __init__(self, unit: int):
        """L2 Normalized `Dense` layer usually used as output layer.

        Args:
            unit: Dimension of the output embbeding. Commonly between
            32 and 512. Larger embeddings usually result in higher accuracy up
            to a point at the expense of making search slower.
        """
        self.unit = unit
        self.dense = layers.Dense(unit)
        super().__init__()
        # FIXME: enforce the shape
        # self.input_spec = rank2

    def call(self, inputs: FloatTensor) -> FloatTensor:
        x = self.dense(inputs)
        normed_x: FloatTensor = tf.math.l2_normalize(x, axis=1)
        return normed_x

    def get_config(self) -> Dict[str, int]:
        return {'unit': self.unit}



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Metrics computed over embeddings"
import tensorflow as tf
from tensorflow.keras.metrics import Metric

from tensorflow_similarity.algebra import build_masks
from tensorflow_similarity.algebra import masked_min, masked_max
from tensorflow_similarity.distances import distance_canonicalizer


@tf.keras.utils.register_keras_serializable(package="Similarity")
class DistanceMetric(Metric):
    def __init__(self,
                 distance,
                 aggregate='mean',
                 anchor="positive",
                 name=None,
                 positive_mining_strategy="hard",
                 negative_mining_strategy="hard",
                 **kwargs):

        if not name:
            name = "%s_%s" % (aggregate, anchor[:3])
        super().__init__(name=name, **kwargs)

        self.distance = distance_canonicalizer(distance)

        if anchor not in ['positive', 'negative']:
            raise ValueError('Invalid anchor_type')
        self.anchor = anchor

        if positive_mining_strategy not in ['hard', 'semi-hard', 'easy']:
            raise ValueError('Invalid positive_mining_strategy')
        self.positive_mining_strategy = positive_mining_strategy

        if negative_mining_strategy not in ['hard', 'easy']:
            raise ValueError('Invalid positive_mining_strategy')
        self.negative_mining_strategy = negative_mining_strategy

        if aggregate not in ['mean', 'max', 'avg', 'min', 'sum']:
            raise ValueError('Invalid reduction')
        self.aggregate = aggregate

        # result variable
        self.aggregated_distances = tf.Variable(0, dtype='float32')

    def update_state(self, labels, embeddings, sample_weight):

        # [distances]
        pairwise_distances = self.distance(embeddings)

        # [mask]
        batch_size = tf.size(labels)
        positive_mask, negative_mask = build_masks(labels, batch_size)

        if self.anchor == "positive":
            if self.positive_mining_strategy == "hard":
                distances, _ = masked_max(pairwise_distances, positive_mask)
            else:
                distances, _ = masked_min(pairwise_distances, positive_mask)
        else:
            if self.negative_mining_strategy == 'easy':
                distances, _ = masked_min(pairwise_distances, negative_mask)
            else:
                distances, _ = masked_max(pairwise_distances, negative_mask)

        # reduce
        if self.aggregate == 'mean' or self.aggregate == 'avg':
            aggregated_distances = tf.reduce_mean(distances)
        elif self.aggregate == 'max':
            aggregated_distances = tf.reduce_max(distances)
        elif self.aggregate == 'min':
            aggregated_distances = tf.reduce_min(distances)
        elif self.aggregate == 'sum':
            aggregated_distances = tf.reduce_sum(distances)

        self.aggregated_distances = aggregated_distances

    def reset_state(self):
        self.aggregated_distances = tf.Variable(0, dtype='float32')

    def result(self):
        return self.aggregated_distances

    def get_config(self):
        config =  {
            "distance": self.distance.name,
            "aggregate": self.aggregate,
            "anchor": self.anchor,
            "positive_mining_strategy": self.positive_mining_strategy,
            "negative_mining_strategy": self.negative_mining_strategy
        }
        base_config = super().get_config()
        return {**base_config, **config}


@tf.keras.utils.register_keras_serializable(package="Similarity")
class DistanceGapMetric(Metric):
    def __init__(self,
                 distance,
                 name=None,
                 **kwargs):
        name = name if name else 'dist_gap'
        super().__init__(name=name, **kwargs)
        self.distance = distance
        self.max_pos_fn = DistanceMetric(distance, aggregate='max')
        self.min_neg_fn = DistanceMetric(distance,
                                         aggregate='min',
                                         anchor='negative')
        self.gap = tf.Variable(0, dtype='float32')

    def update_state(self, labels, embeddings, sample_weight):
        max_pos = self.max_pos_fn(labels, embeddings, sample_weight)
        min_neg = self.min_neg_fn(labels, embeddings, sample_weight)
        self.gap.assign(tf.abs(min_neg - max_pos))

    def result(self):
        return self.gap

    def get_config(self):
        config = {
            "distance": self.distance,
        }
        base_config = super().get_config()
        return {**base_config, **config}


# aliases
@tf.keras.utils.register_keras_serializable(package="Similarity")
def dist_gap(distance):
    return DistanceGapMetric(distance)


# max
@tf.keras.utils.register_keras_serializable(package="Similarity")
def max_pos(distance):
    return DistanceMetric(distance, aggregate='max')


@tf.keras.utils.register_keras_serializable(package="Similarity")
def max_neg(distance):
    return DistanceMetric(distance, aggregate='max', anchor='negative')


# avg
@tf.keras.utils.register_keras_serializable(package="Similarity")
def avg_pos(distance):
    return DistanceMetric(distance, aggregate='mean')


@tf.keras.utils.register_keras_serializable(package="Similarity")
def avg_neg(distance):
    return DistanceMetric(distance, aggregate='mean', anchor='negative')


# min
@tf.keras.utils.register_keras_serializable(package="Similarity")
def min_pos(distance):
    return DistanceMetric(distance, aggregate='min')


@tf.keras.utils.register_keras_serializable(package="Similarity")
def min_neg(distance):
    return DistanceMetric(distance, aggregate='min', anchor='negative')


# sum
@tf.keras.utils.register_keras_serializable(package="Similarity")
def sum_pos(distance):
    return DistanceMetric(distance, aggregate='sum')


@tf.keras.utils.register_keras_serializable(package="Similarity")
def sum_neg(distance):
    return DistanceMetric(distance, aggregate='sum', anchor='negative')



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Efficiently find nearest indexed embeddings

The search module is used to find the closest indexed example embeddings
to a query example embebbeding.
To do so it performs a sub-linear time
[ANN (Approximate Nearst Neigboors)](https://en.wikipedia.org/wiki/Nearest_neighbor_search)
search on the indexed set of embedding examples.

Different ANN librairies have [different performance profiles](https://github.com/erikbern/ann-benchmarks).
Tensorflow Similarity by default use [NMSLIB](https://github.com/nmslib/nmslib)
which has a strong performance profile and is very portable.

Adding another backend is fairly straightforward: subclass the
abstract class `Search()` and implement the abstract methods. Then to use it
pass it to the `compile()` method of your [SimilarityModel].

Adding your search backend as a built-in choice invlolves
modifiying the [Indexer](../indexer.md) and sending a PR. In general, unless
the backend is of general use, its better to not include it as
a built-in option as it must be supported moving forward.
"""
from .nmslib_search import NMSLibSearch # noqa
from .search import Search  # noqa



import tensorflow as tf
import numpy as np
from tensorflow_similarity.losses import TripletLoss
from tensorflow_similarity.losses import PNLoss


# [triplet loss]
def test_triplet_loss_serialization():
    loss = TripletLoss()
    config = loss.get_config()
    print(config)
    loss2 = TripletLoss.from_config(config)
    assert loss.name == loss2.name
    assert loss.distance == loss2.distance


def triplet_hard_loss_np(labels, embedding, margin, dist_func, soft=False):

    num_data = embedding.shape[0]
    # Reshape labels to compute adjacency matrix.
    labels_reshaped = np.reshape(labels.astype(np.float32),
                                 (labels.shape[0], 1))

    adjacency = np.equal(labels_reshaped, labels_reshaped.T)
    pdist_matrix = dist_func(embedding)
    loss_np = 0.0
    for i in range(num_data):
        pos_distances = []
        neg_distances = []
        for j in range(num_data):
            if adjacency[i][j] == 0:
                neg_distances.append(pdist_matrix[i][j])
            if adjacency[i][j] > 0.0 and i != j:
                pos_distances.append(pdist_matrix[i][j])

        # if there are no positive pairs, distance is 0
        if len(pos_distances) == 0:
            pos_distances.append(0)

        # Sort by distance.
        neg_distances.sort()
        min_neg_distance = neg_distances[0]
        pos_distances.sort(reverse=True)
        max_pos_distance = pos_distances[0]

        if soft:
            loss_np += np.log1p(np.exp(max_pos_distance - min_neg_distance))
        else:
            loss_np += np.maximum(0.0,
                                  max_pos_distance - min_neg_distance + margin)

    loss_np /= num_data
    return loss_np


def test_triplet_loss():
    num_inputs = 10
    # y_true: labels
    y_true = tf.random.uniform((num_inputs, ), 0, 10, dtype=tf.int32)
    # y_preds: embedding
    y_preds = tf.random.uniform((num_inputs, 20), 0, 1)
    tpl = TripletLoss()
    # y_true, y_preds
    loss = tpl(y_true, y_preds)
    assert loss > 0.9


def test_triplet_loss_easy():
    num_inputs = 10
    # y_true: labels
    y_true = tf.random.uniform((num_inputs, ), 0, 3, dtype=tf.int32)
    # y_preds: embedding
    y_preds = tf.random.uniform((num_inputs, 16), 0, 1)
    tpl = TripletLoss(positive_mining_strategy='easy',
                      negative_mining_strategy='easy')
    # y_true, y_preds
    loss = tpl(y_true, y_preds)
    assert loss > 0


def test_triplet_loss_semi_hard():
    num_inputs = 10
    # y_true: labels
    y_true = tf.random.uniform((num_inputs, ), 0, 3, dtype=tf.int32)
    # y_preds: embedding
    y_preds = tf.random.uniform((num_inputs, 16), 0, 1)
    tpl = TripletLoss(positive_mining_strategy='easy',
                      negative_mining_strategy='semi-hard')
    # y_true, y_preds
    loss = tpl(y_true, y_preds)
    assert loss


def test_triplet_loss_hard():
    num_inputs = 10
    # y_true: labels
    y_true = tf.random.uniform((num_inputs, ), 0, 3, dtype=tf.int32)
    # y_preds: embedding
    y_preds = tf.random.uniform((num_inputs, 16), 0, 1)
    tpl = TripletLoss(positive_mining_strategy='hard',
                      negative_mining_strategy='hard')
    # y_true, y_preds
    loss = tpl(y_true, y_preds)
    assert loss


# [pn loss]
def test_pn_loss_serialization():
    loss = PNLoss()
    config = loss.get_config()
    print(config)
    loss2 = PNLoss.from_config(config)
    assert loss.name == loss2.name
    assert loss.distance == loss2.distance


def test_np_loss():
    num_inputs = 10
    # y_true: labels
    y_true = tf.random.uniform((num_inputs, ), 0, 10, dtype=tf.int32)
    # y_preds: embedding
    y_preds = tf.random.uniform((num_inputs, 20), 0, 1)
    pnl = PNLoss()
    # y_true, y_preds
    loss = pnl(y_true, y_preds)
    assert loss > 0.9



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf

from tensorflow_similarity.types import IntTensor


def confusion_matrix(y_pred: IntTensor,
                     y_true: IntTensor,
                     normalize: bool = True,
                     labels: IntTensor = None,
                     title: str = 'Confusion matrix',
                     cmap: str = 'Blues'):
    """Plot confusion matrix

    Args:
        y_pred: Model prediction returned by `model.match()`

        y_true: Expected class_id.

        normalize: Normalizes matrix values between 0 and 1.
        Defaults to True.

        labels: List of class string label to display instead of the class
        numerical ids. Defaults to None.

        title: Title of the confusion matrix. Defaults to 'Confusion matrix'.

        cmap: Color schema as CMAP. Defaults to 'Blues'.
    """
    # Ensure we are working with integer tensors.
    y_pred = tf.cast(tf.convert_to_tensor(y_pred), dtype='int32')
    y_true = tf.cast(tf.convert_to_tensor(y_true), dtype='int32')

    cm = tf.math.confusion_matrix(y_true, y_pred)
    cm = tf.cast(cm, dtype='float')
    accuracy = tf.linalg.trace(cm) / tf.math.reduce_sum(cm)
    misclass = 1 - accuracy

    if normalize:
        cm = tf.math.divide_no_nan(
                cm,
                tf.math.reduce_sum(cm, axis=1)[:, np.newaxis]
        )

    plt.figure(figsize=(8, 6))
    plt.imshow(cm, interpolation='nearest', cmap=cmap)
    plt.title(title)
    plt.colorbar()

    if labels is not None:
        tick_marks = np.arange(len(labels))
        plt.xticks(tick_marks, labels, rotation=45)
        plt.yticks(tick_marks, labels)

    cm_max = tf.math.reduce_max(cm)
    thresh = cm_max / 1.5 if normalize else cm_max / 2.0
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        val = cm[i, j]
        color = "white" if val > thresh else "black"
        txt = "%.2f" % val if val > 0.0 else "0"
        plt.text(j, i, txt, horizontalalignment="center", color=color)

    plt.tight_layout()
    plt.ylabel('True label')
    plt.xlabel('Predicted label\naccuracy={:0.4f}; misclass={:0.4f}'.format(
        accuracy, misclass))
    plt.show()



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Data Samplers generate balanced batches for smooth training.

*A well balanced batch is a batch that contains at least 2 examples for
each class present in the batch*.

Having well balanced batches is important for many types of similarity learning
including contrastive learning because contrastive losses require at least
two examples (and sometimes more) to be able to compute distances between
the embeddings.

To address this need, TensorFlow Similarity provides data samplers for
various types of datasets that:
- Ensure that batches contain at least N examples of each class present in
the batch.
- Support restricting the batches to a subset of the classes present in
the dataset.
"""
from .utils import select_examples  # noqa
from .memory_samplers import MultiShotMemorySampler  # noqa
from .memory_samplers import SingleShotMemorySampler  # noqa
from .tfrecords_samplers import TFRecordDatasetSampler  # noqa
from .tfdataset_samplers import TFDatasetMultiShotMemorySampler  # noqa



import numpy as np
import tensorflow as tf

from tensorflow_similarity.retrieval_metrics import MapAtK


def test_concrete_instance():
    rm = MapAtK(r={1: 4, 0: 4})

    assert rm.name == "map@5"
    # Check the name once we have updated the threshold
    rm.distance_threshold = 0.1
    assert rm.name == "map@5 : distance_threshold@0.1"
    assert repr(rm) == "map@k : map@5 : distance_threshold@0.1"
    assert rm.canonical_name == "map@k"
    assert rm.k == 5
    assert rm.distance_threshold == 0.1
    assert rm.average == "micro"

    expected_config = {
        "r": {1: 4, 0: 4},
        "name": "map@5 : distance_threshold@0.1",
        "canonical_name": "map@k",
        "k": 5,
        "distance_threshold": 0.1,
    }
    assert rm.get_config() == expected_config


def test_compute():
    query_labels = tf.constant([1, 1, 1, 0])
    match_mask = tf.constant(
        [
            [True, True, False],
            [True, True, False],
            [True, True, False],
            [False, False, True],
        ],
        dtype=bool,
    )
    rm = MapAtK(r={0: 10, 1: 3},  k=3)

    mapk = rm.compute(query_labels=query_labels, match_mask=match_mask)

    # mapk should be sum(precision@k*Relevancy_Mask)/R
    # class 1 has 3 results sets that are all T,T,F:
    #     (1.0*True+1.0*True+0.66*False)/3 = 0.66667
    # class 0 has 1 result set that is F,F,T
    #     (0.0*False+0.0*False+0.33*True)/10 = 0.03332
    # mapk = (0.667*3 + 0.0332)/4
    expected = tf.constant(0.50833333332)

    np.testing.assert_allclose(mapk, expected)



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Callable, Optional

import tensorflow as tf


def TFRecordDatasetSampler(shard_path: str,
                           deserialization_fn: Callable,
                           example_per_class: int = 2,
                           batch_size: int = 32,
                           shards_per_cycle: int = None,
                           compression: Optional[str] = None,
                           parallelism: int = tf.data.AUTOTUNE,
                           file_parallelism: int = 1,
                           prefetch_size: Optional[int] = None,
                           shard_suffix: str = "*.tfrec") -> tf.data.Dataset:
    """Create a [TFRecordDataset](https://www.tensorflow.org/api_docs/python/tf/data/TFRecordDataset)
    based sampler

    This sampler should be used when using a TFDataset or have a large
    dataset that needs to be stored on file.

    **WARNING**: This samplers assume that classes examples are contigious,
    at least enough that you can get `example_per_class` numbers
    of them consecutively. This requirements is needed to make the
    sampling efficient and makes dataset constuctionn oftentime easier as
    there is no need to worry about shuffling. Somewhat contigious means
    its fine to have the same class in multiples shards as long as the
    examples for the same classes are contigious in that shard.

    Overall the way we make the sampling process is by using the
    [tf.dataset.interleaves](https://www.tensorflow.org/api_docs/python/tf/data/Dataset#interleave)
    in a non orthodox way: we use its `block_length` to control the
    number of example per class and rely on the parallelize &
    non_deterministic version of the API to do the sampling efficiently
    for us. Relying on pure tf.data ops also ensure good compatibility with
    distribution strategy.


    Args:
        shard_path: Directory where the shards are stored.

        deserialization_fn: Function used to deserialize the tfRecord and
        construct a valid example.

        example_per_class: Number of example per class in each batch.
        Defaults to 2.

        batch_size: How many examples in each batch. The number of class in
        the batch will be `batch_size // example_per_class`.
        Defaults to 32.

        shards_per_cycle: How many shards to use concurrently per cycle.
        Default is None which is all of them. Can cause segv if too many
        shards.

        compression: Which compression was used when creating the dataset.
        `{None, "ZLIB", or "GZIP"}` as specified in [TFRecordDataset
        documentation](https://www.tensorflow.org/api_docs/python/tf/data/TFRecordDataset)
        Defaults to None.

        parallelism: How many parallel calls to do. If not set, will let
        TensorFlow decide by using `tf.data.AUTOTUNE` (-1).

        file_parallelism: How many parallel shards to read increase number
        if IO bound. Defaults to 1.

        prefetch_size: How many batch to precache. Defaults to 10.

        shard_suffix: Glog pattern used to collect the shard files list.
        Defaults to "*.tfrec".

    Returns:
        A `TF.data.dataset` ready to be consumed by the model.
    """
    shards_list = [str(i) for i in Path(shard_path).glob(shard_suffix)]
    total_shards = len(shards_list)
    print(f'found {total_shards} shards')

    if not prefetch_size:
        prefetch_size = 10

    # how many shard to iterate over in parallels.
    cycle_length = shards_per_cycle if shards_per_cycle else total_shards

    with tf.device('/cpu:0'):
        # shuffle the shard order
        ds = tf.data.Dataset.from_tensor_slices(shards_list)

        # shuffle shard order
        ds = ds.shuffle(total_shards)

        # This is the tricky part, we are using the interleave function to
        # do the sampling as requested by the user. This is not the
        # standard use of the function or an obvious way to do it but
        # its by far the faster and more compatible way to do so
        # we are favoring for once those factors over readability
        # deterministic=False is not an error, it is what allows us to
        # create random batch
        ds = ds.interleave(
            lambda x: tf.data.TFRecordDataset(
                x,
                compression_type=compression,
            ),  # noqa
            cycle_length=cycle_length,
            block_length=example_per_class,
            num_parallel_calls=file_parallelism,
            deterministic=False,
        )
        ds = ds.map(deserialization_fn, num_parallel_calls=parallelism)
        ds = ds.batch(batch_size)
        ds = ds.repeat()
        ds = ds.prefetch(prefetch_size)
        return ds



import tensorflow as tf
from tensorflow_similarity.losses import TripletLoss
from tensorflow_similarity.models import SimilarityModel


def test_save_and_reload(tmp_path):
    inputs = tf.keras.layers.Input(shape=(3,))
    outputs = tf.keras.layers.Dense(2)(inputs)
    model = SimilarityModel(inputs, outputs)
    model.compile(optimizer='adam', loss=TripletLoss())

    # index data
    x = tf.constant([[1, 1, 3], [3, 1, 2]], dtype='float32')
    y = tf.constant([1, 2])
    model.index(x, y)

    # save
    model.save(tmp_path)

    # reload
    loaded_model = tf.keras.models.load_model(tmp_path)
    loaded_model.load_index(tmp_path)
    assert loaded_model._index.size() == len(y)


def test_save_no_compile(tmp_path):

    inputs = tf.keras.layers.Input(shape=(3,))
    outputs = tf.keras.layers.Dense(2)(inputs)
    model = SimilarityModel(inputs, outputs)

    model.save(tmp_path)
    model2 = tf.keras.models.load_model(tmp_path)
    assert isinstance(model2, type(model))



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from tensorflow_similarity.samplers.samplers import Sampler

import tensorflow as tf


def batch_class_ratio(sampler: Sampler, num_batches: int = 100) -> float:
    """Computes the average number of examples per class within each batch.
    Similarity learning requires at least 2 examples per class in each batch.
    This is needed in order to construct the triplets. This function
    provides the average number of examples per class within each batch and
    can be used to check that a sampler is working correctly.
    The ratio should be >= 2.
    Args:
        sampler: A tf.similarity sampler object.
        num_batches: The number of batches to sample.
    Returns:
        The average number of examples per class.
    """
    ratio = 0
    for batch_count, (_, y) in enumerate(sampler):
        if batch_count < num_batches:
            batch_size = tf.shape(y)[0]
            num_classes = tf.shape(tf.unique(y)[0])[0]
            ratio += tf.math.divide(batch_size, num_classes)
        else:
            break

    return float(ratio/(batch_count+1))



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from tensorflow_similarity.types import FloatTensor
from .classification_metric import ClassificationMetric


class F1Score(ClassificationMetric):
    """Calculates the harmonic mean of precision and recall.

    Computes the F-1 Score given the query classification counts. The metric is
    computed as follows:

    $$
    F_1 = 2 \cdot \frac{\textrm{precision} \cdot \textrm{recall}}{\textrm{precision} + \textrm{recall}}
    $$

    args:
        name: Name associated with a specific metric object, e.g.,
        f1@0.1

    Usage with `tf.similarity.models.SimilarityModel()`:

    ```python
    model.calibrate(x=query_examples,
                    y=query_labels,
                    calibration_metric='f1')
    ```
    """
    def __init__(self, name: str = 'f1') -> None:
        super().__init__(
                name=name,
                canonical_name='f1_score')

    def compute(self,
                tp: FloatTensor,
                fp: FloatTensor,
                tn: FloatTensor,
                fn: FloatTensor,
                count: int) -> FloatTensor:
        """Compute the classification metric.

        The `compute()` method supports computing the metric for a set of
        values, where each value represents the counts at a specific distance
        threshold.

        Args:
            tp: A 1D FloatTensor containing the count of True Positives at each
            distance threshold.

            fp: A 1D FloatTensor containing the count of False Positives at
            each distance threshold.

            tn: A 1D FloatTensor containing the count of True Negatives at each
            distance threshold.

            fn: A 1D FloatTensor containing the count of False Negatives at
            each distance threshold.

            count: The total number of queries

        Returns:
            A 1D FloatTensor containing the metric at each distance threshold.
        """
        recall = tf.math.divide_no_nan(tp, tp + fn)
        precision = tf.math.divide_no_nan(tp, tp + fp)
        numer = 2 * recall * precision
        denom = recall + precision
        result: FloatTensor = tf.math.divide_no_nan(numer, denom)
        return result



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Mapping, Optional, Sequence, Tuple

from matplotlib import pyplot as plt

from tensorflow_similarity.types import Tensor, Lookup


def viz_neigbors_imgs(example: Tensor,
                      example_class: int,
                      neighbors: Sequence[Lookup],
                      class_mapping: Optional[Mapping[int, str]] = None,
                      fig_size: Tuple[int, int] = (24, 4),
                      cmap: str = 'viridis'):
    """Display images nearest neighboors

    Args:
        example: The data used as query input.

        example_class: The class of the data used as query

        neighbors: The list of neighbors returned by the lookup()

        class_mapping: Dictionary that map the class numerical ids to a class
        name. If not set, will display the class numerical id.
        Defaults to None.

        fig_size: Size of the figure. Defaults to (24, 4).

        cmap: Default color scheme for black and white images e.g mnist.
        Defaults to 'viridis'.
    """
    num_cols = len(neighbors) + 1
    plt.subplots(1, num_cols, figsize=fig_size)
    plt_idx = 1

    # draw target
    plt.subplot(1, num_cols, plt_idx)
    plt.imshow(example, cmap=cmap)
    plt.xticks([])
    plt.yticks([])

    val = class_mapping[example_class] if class_mapping else str(example_class)
    plt.title(val)
    plt_idx += 1

    for nbg in neighbors:
        plt.subplot(1, num_cols, plt_idx)
        if class_mapping and nbg.label is not None:
            val = class_mapping[nbg.label]
        elif nbg.label is not None:
            val = str(nbg.label)
        else:
            val = 'No Label'
        legend = f"{val} - {nbg.distance:.5f}"
        if nbg.label == example_class:
            color = cmap
        else:
            color = 'Reds'
        plt.imshow(nbg.data, cmap=color)
        plt.title(legend)
        plt.xticks([])
        plt.yticks([])

        plt_idx += 1
    plt.show()



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from .retrieval_metric import RetrievalMetric
from tensorflow_similarity.types import FloatTensor, IntTensor, BoolTensor


class RecallAtK(RetrievalMetric):
    """The metric learning version of Recall@K.

    A query is counted as a positive when ANY lookup in top K match the query
    class, 0 otherwise.

    Args:
        name: Name associated with the metric object, e.g., recall@5

        canonical_name: The canonical name associated with metric,
        e.g., recall@K

        k: The number of nearest neighbors over which the metric is computed.

        distance_threshold: The max distance below which a nearest neighbor is
        considered a valid match.

        average: {'micro'} Determines the type of averaging performed over the
        queries.

        * 'micro': Calculates metrics globally over all queries.

        * 'macro': Calculates metrics for each label and takes the unweighted
                   mean.
    """

    def __init__(self, name: str = "recall", k: int = 5, **kwargs) -> None:
        if "canonical_name" not in kwargs:
            kwargs["canonical_name"] = "recall@k"

        super().__init__(name=name, k=k, **kwargs)

    def compute(
        self,
        *,  # keyword only arguments see PEP-570
        query_labels: IntTensor,
        match_mask: BoolTensor,
        **kwargs,
    ) -> FloatTensor:
        """Compute the metric

        Args:
            query_labels: A 1D tensor of the labels associated with the
            embedding queries.

            match_mask: A 2D mask where a 1 indicates a match between the
            jth query and the kth neighbor and a 0 indicates a mismatch.

            **kwargs: Additional compute args.

        Returns:
            A rank 0 tensor containing the metric.
        """
        self._check_shape(query_labels, match_mask)

        k_slice = match_mask[:, : self.k]
        match_indicator = tf.math.reduce_any(k_slice, axis=1)
        match_indicator = tf.cast(match_indicator, dtype="float")

        if self.average == "micro":
            recall_at_k = tf.math.reduce_mean(match_indicator)
        elif self.average == "macro":
            per_class_metrics = 0
            class_labels = tf.unique(query_labels)[0]
            # TODO(ovallis): potential slowness.
            for label in class_labels:
                idxs = tf.where(query_labels == label)
                c_slice = tf.gather(match_indicator, indices=idxs)
                per_class_metrics += tf.math.reduce_mean(c_slice)
            recall_at_k = tf.math.divide(per_class_metrics, len(class_labels))
        else:
            raise ValueError(
                f"{self.average} is not a supported average " "option"
            )
        result: FloatTensor = recall_at_k
        return result



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod

from typing import List, Optional, Sequence, Tuple

from tensorflow_similarity.types import FloatTensor, PandasDataFrame, Tensor


class Store(ABC):

    @abstractmethod
    def add(self,
            embedding: FloatTensor,
            label: Optional[int] = None,
            data: Optional[Tensor] = None) -> int:
        """Add an Embedding record to the key value store.

        Args:
            embedding: Embedding predicted by the model.

            label: Class numerical id. Defaults to None.

            data: Data associated with the embedding. Defaults to None.

        Returns:
            Associated record id.
        """
    @abstractmethod
    def batch_add(self,
                  embeddings: Sequence[FloatTensor],
                  labels: Optional[Sequence[int]] = None,
                  data: Optional[Sequence[Tensor]] = None) -> List[int]:
        """Add a set of embedding records to the key value store.

        Args:
            embeddings: Embeddings predicted by the model.

            labels: Class numerical ids. Defaults to None.

            data: Data associated with the embeddings. Defaults to None.

        See:
            add() for what a record contains.

        Returns:
            List of associated record id.
        """

    @abstractmethod
    def get(self, idx: int) -> Tuple[FloatTensor,
                                     Optional[int],
                                     Optional[Tensor]]:
        """Get an embedding record from the key value store.

        Args:
            idx: Id of the record to fetch.

        Returns:
            record associated with the requested id.
        """

    @abstractmethod
    def batch_get(self,
                  idxs: Sequence[int]) -> Tuple[List[FloatTensor],
                                                List[Optional[int]],
                                                List[Optional[Tensor]]]:
        """Get embedding records from the key value store.

        Args:
            idxs: ids of the records to fetch.

        Returns:
            List of records associated with the requested ids.
        """

    @abstractmethod
    def size(self) -> int:
        "Number of record in the key value store."

    @abstractmethod
    def save(self, path: str, compression: bool = True) -> None:
        """Serializes index on disk.

        Args:
            path: Directory where to store the data.
            compression: Compress index data. Defaults to True.
        """

    @abstractmethod
    def load(self, path: str) -> int:
        """Load index on disk

        Args:
            path: where to store the data

        Returns:
           Number of records reloaded.
        """

    @abstractmethod
    def to_data_frame(self, num_records: int = 0) -> PandasDataFrame:
        """Export data as a Pandas dataframe.

        Args:
            num_records: Number of records to export to the dataframe.
            Defaults to 0 (unlimited).

        Returns:
            pd.DataFrame: a pandas dataframe.
        """



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Circle Loss
    Circle Loss: A Unified Perspective of Pair Similarity Optimization
    https://arxiv.org/abs/2002.10857
"""

import tensorflow as tf
from typing import Any, Callable, Union

from tensorflow_similarity.distances import Distance, distance_canonicalizer
from tensorflow_similarity.algebra import build_masks
from tensorflow_similarity.types import FloatTensor, IntTensor
from .metric_loss import MetricLoss
from .utils import logsumexp


@tf.keras.utils.register_keras_serializable(package="Similarity")
@tf.function
def circle_loss(labels: IntTensor,
                embeddings: FloatTensor,
                distance: Callable,
                gamma: float = 80,
                margin: float = 0.4) -> Any:
    """Circle loss computations

    The original paper used cosine similarity while this loss has been modified
    to work with cosine distance.

    Args:
        labels: Labels associated with the embeddings

        embeddings: Embeddings as infered by the model.

        distance: Which distance function to use to compute the pairwise
        distances between embeddings. The distance is expected to be
        between [0, 2]. Defaults to 'cosine'.

        gamma: Scaling term. Defaults to 80. Note: Large values cause the
        LogSumExp to return the Max pair and reduces the weighted mixing of all
        pairs. Should be hypertuned.

        margin: Used to weight the distance. Below this distance, negatives are
        up weighted and positives are down weighted. Similarly, above this
        distance negatives are down weighted and positive are up weighted.
        Defaults to 0.4.

    Returns:
        Loss: The loss value for the current batch.
    """

    # Switched from what's in the paper to work with distance instead of
    # similarity.
    optim_pos = margin
    optim_neg = 1 + margin
    delta_pos = margin
    delta_neg = 1 - margin

    # label
    batch_size = tf.size(labels)

    # [distances]
    pairwise_distances = distance(embeddings)

    # [masks] -> filter to keep only the relevant value - zero the rest
    positive_mask, negative_mask = build_masks(labels, batch_size)
    valid_anchors = tf.math.logical_and(
            tf.math.reduce_any(positive_mask, axis=1),
            tf.math.reduce_any(negative_mask, axis=1)
    )

    # Cast masks as floats to support multiply
    valid_anchors = tf.cast(valid_anchors, dtype='float32')
    positive_mask = tf.cast(positive_mask, dtype='float32')
    negative_mask = tf.cast(negative_mask, dtype='float32')

    # [weights] from  (5) in 3.1 using optim values of 3.2
    # Implementation note: we do all the computation on the full pairwise and
    # filter at then end to keep only relevant values.

    # positive weights
    pos_weights = optim_pos + pairwise_distances  # (5) in 3.1
    pos_weights = pos_weights * positive_mask  # filter
    pos_weights = tf.maximum(pos_weights, 0.0)  # clip at zero

    # negative weights
    neg_weights = optim_neg - pairwise_distances  # (5) in 3.1
    neg_weights = neg_weights * negative_mask  # filter
    neg_weights = tf.maximum(neg_weights, 0.0)  # clip at zero

    # Subtract the between and within class margins
    pos_dists = delta_pos - pairwise_distances
    neg_dists = delta_neg - pairwise_distances

    # distances filtering
    # /2 because we have a pairwise so each distance is counted twice
    # applying weights as in (4) in 3.1
    pos_wdists = (-1 * gamma * pos_weights * pos_dists)  # / 2
    neg_wdists = (gamma * neg_weights * neg_dists)  # / 2

    p_loss = logsumexp(pos_wdists, positive_mask)
    n_loss = logsumexp(neg_wdists, negative_mask)

    # Remove any anchors that have empty neg or pos pairs.
    # NOTE: reshape is required here because valid_anchors is [m] and
    #       p_loss + n_loss is [m, 1].
    circle_loss = tf.math.multiply(
            p_loss + n_loss,
            tf.reshape(valid_anchors, (-1, 1))
    )
    # reduce and scale loss so it isn't a function of the batch size.
    circle_loss = tf.math.reduce_mean(circle_loss)

    return circle_loss


@tf.keras.utils.register_keras_serializable(package="Similarity")
class CircleLoss(MetricLoss):
    """Computes the CircleLoss

    Circle Loss: A Unified Perspective of Pair Similarity Optimization
    https://arxiv.org/abs/2002.10857

    The original paper used cosine similarity while this loss has been
    modified to work with cosine distance.

    `y_true` must be  a 1-D integer `Tensor` of shape (batch_size,).
    It's values represent the classes associated with the examples as
    **integer  values**.

    `y_pred` must be 2-D float `Tensor`  of L2 normalized embedding vectors.
    you can use the layer `tensorflow_similarity.layers.L2Embedding()` as the
    last layer of your model to ensure your model output is properly
    normalized.
    """
    def __init__(self,
                 distance: Union[Distance, str] = 'cosine',
                 gamma: float = 80.0,
                 margin: float = 0.40,
                 name: str = 'CircleLoss'):
        """Initializes a CircleLoss

        Args:
            distance: Which distance function to use to compute the pairwise
            distances between embeddings. The distance is expected to be
            between [0, 2]. Defaults to 'cosine'.

            gamma: Scaling term. Defaults to 80. Note: Large values cause the
            LogSumExp to return the Max pair and reduces the weighted mixing
            of all pairs. Should be hypertuned.

            margin: Used to weight the distance. Below this distance, negatives
            are up weighted and positives are down weighted. Similarly, above
            this distance negatives are down weighted and positive are up
            weighted. Defaults to 0.4.

            name: Loss name. Defaults to CircleLoss.
        """

        # distance canonicalization
        distance = distance_canonicalizer(distance)
        self.distance = distance

        super().__init__(circle_loss,
                         name=name,
                         reduction=tf.keras.losses.Reduction.NONE,
                         distance=distance,
                         gamma=gamma,
                         margin=margin)



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Mapping

import tensorflow as tf

from .retrieval_metric import RetrievalMetric
from tensorflow_similarity.types import FloatTensor, IntTensor, BoolTensor


class MapAtK(RetrievalMetric):
    """Mean Average precision - mAP@K is computed as.

    $$
    mAP_i@K = \frac{\sum_{j = 1}^{K} {rel_i_j}\times{P_i@j}}{R}
    $$

    Where: K is the number of neighbors in the i_th query result set.
           P is the rolling precision over the i_th query result set.
           R is the cardinality of the target class.
           rel is the relevance mask (indicator function) for the i_th query.
           i represents the i_th query.
           j represents the j_th ranked query result.

    AP@K is biased towards the top ranked results and is a function of the rank
    (K), the relevancy mask (rel), and the number of indexed examples for the
    class (R). The denominator for the i_th query is set to the number of
    indexed examples (R) for the class associated with the i_th query.

    For example, if the index has has 100 embedded examples (R) of class 'a',
    and our query returns 50 results (K) where the top 10 results are all TPs,
    then the AP@50 will be 0.10; however, if instead the bottom 10 ranked
    results are all TPs, then the AP@50 will be much lower (0.012) because we
    apply a penalty for the 40 FPs that come before the relevant query results.

    This metric is useful when we want to ensure that the top ranked results
    are relevant to the query; however, it requires that we pass a mapping from
    the class id to the number of indexed examples for that class.

    Args:
        r: A mapping from class id to the number of examples in the index,
        e.g., r[4] = 10 represents 10 indexed examples from class 4.

        name: Name associated with the metric object, e.g., avg_precision@5

        canonical_name: The canonical name associated with metric, e.g.,
        avg_precision@K

        k: The number of nearest neighbors over which the metric is computed.

        distance_threshold: The max distance below which a nearest neighbor is
        considered a valid match.

        average: {'micro'} Determines the type of averaging performed over the
        queries.

        * 'micro': Calculates metrics globally over all queries.
    """

    def __init__(
        self,
        r: Mapping[int, int],
        name: str = "map",
        k: int = 5,
        average: str = "micro",
        **kwargs,
    ) -> None:
        if average == "macro":
            raise ValueError(
                "Mean Average Precision only supports micro averaging."
            )

        if "canonical_name" not in kwargs:
            kwargs["canonical_name"] = "map@k"

        super().__init__(name=name, k=k, average=average, **kwargs)
        self.r = r

    def get_config(self):
        config = {
            "r": self.r,
        }
        base_config = super().get_config()
        return {**base_config, **config}

    def compute(
        self,
        *,  # keyword only arguments see PEP-570
        query_labels: IntTensor,
        match_mask: BoolTensor,
        **kwargs,
    ) -> FloatTensor:
        """Compute the metric

        Args:
            query_labels: A 1D array of the labels associated with the
            embedding queries.

            match_mask: A 2D mask where a 1 indicates a match between the
            jth query and the kth neighbor and a 0 indicates a mismatch.

            **kwargs: Additional compute args

        Returns:
            A rank 0 tensor containing the metric.
        """
        self._check_shape(query_labels, match_mask)

        k_slice = tf.cast(match_mask[:, : self.k], dtype="float")
        tp = tf.math.cumsum(k_slice, axis=1)
        p_at_k = tf.math.divide(tp, tf.range(1, self.k + 1, dtype="float"))
        p_at_k = tf.math.multiply(k_slice, p_at_k)

        if self.average == "micro":
            table = tf.lookup.StaticHashTable(
                tf.lookup.KeyValueTensorInitializer(
                    list(self.r.keys()),
                    list(self.r.values()),
                    key_dtype=tf.int32,
                    value_dtype=tf.int32,
                ),
                default_value=-1,
            )
            class_counts = table.lookup(query_labels)
            avg_p_at_k = tf.math.divide(
                tf.math.reduce_sum(p_at_k, axis=1),
                tf.cast(class_counts, dtype="float"),
            )

            avg_p_at_k = tf.math.reduce_mean(avg_p_at_k)
        else:
            raise ValueError(
                f"{self.average} is not a supported average option"
            )

        result: FloatTensor = avg_p_at_k
        return result



import numpy as np
from tensorflow_similarity.search import NMSLibSearch


def test_index_match():
    target = np.array([1, 1, 2], dtype='float32')
    embs = np.array([[1, 1, 3], [3, 1, 2]], dtype='float32')

    search_index = NMSLibSearch('cosine', 3)
    search_index.add(embs[0], 0)
    search_index.add(embs[1], 1)

    idxs, embs = search_index.lookup(target)

    assert len(embs) == 2
    assert list(idxs) == [0, 1]


def test_index_save(tmp_path):
    target = np.array([1, 1, 2], dtype='float32')
    embs = np.array([[1, 1, 3], [3, 1, 2]], dtype='float32')

    search_index = NMSLibSearch('cosine', 3)
    search_index.add(embs[0], 0)
    search_index.add(embs[1], 1)

    idxs, embs = search_index.lookup(target)

    assert len(embs) == 2
    assert list(idxs) == [0, 1]

    search_index.save(tmp_path)

    search_index2 = NMSLibSearch('cosine', 3)
    search_index2.load(tmp_path)

    idxs2, embs2 = search_index.lookup(target)
    assert len(embs2) == 2
    assert list(idxs2) == [0, 1]

    # add more
    search_index2.add(np.array([3.0, 3.0, 3.0]), 3)
    idxs3, embs3 = search_index2.lookup(target)
    assert len(embs3) == 3
    assert list(idxs3) == [0, 3, 1]


def test_batch_vs_single(tmp_path):
    num_targets = 10
    index_size = 100
    vect_dim = 16

    # gen
    idxs = [i for i in range(index_size)]

    targets = np.random.random((num_targets, vect_dim)).astype('float32')
    embs = np.random.random((index_size, vect_dim)).astype('float32')

    # build search_index
    search_index = NMSLibSearch('cosine', vect_dim)
    search_index.batch_add(embs, idxs)

    # batch
    batch_idxs, _ = search_index.batch_lookup(targets)

    # single
    singles_idxs = []
    for t in targets:
        idxs, embs = search_index.lookup(t)
        singles_idxs.append(idxs)

    for i in range(num_targets):
        # k neigboors are the same?
        for k in range(3):
            assert batch_idxs[i][k] == singles_idxs[i][k]



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Set of useful algebraic functions used through the package"
from typing import Tuple
import tensorflow as tf
from .types import BoolTensor, FloatTensor, IntTensor


def masked_max(distances: FloatTensor,
               mask: BoolTensor,
               dim: int = 1) -> Tuple[FloatTensor, FloatTensor]:
    """Computes the maximum values over masked pairwise distances.

    We need to use this formula to make sure all values are >=0.

    Args:
      distances: 2-D float `Tensor` of [n, n] pairwise distances
      mask: 2-D Boolean `Tensor` of [n, n] valid distance size.
      dim: The dimension over which to compute the maximum. Defaults to 1.

    Returns:
      A Tuple of Tensors containing the maximum distance value and the arg_max
      for each example.
    """
    # Convert to dbl to avoid precision error in offset
    distances = tf.cast(distances, dtype=tf.float64)
    mask = tf.cast(mask, dtype=tf.float64)

    axis_min = tf.math.reduce_min(distances, dim, keepdims=True) - 1e-6
    masked_max = tf.math.multiply(distances - axis_min, mask)
    arg_max = tf.math.argmax(masked_max, dim)
    masked_max = tf.math.reduce_max(masked_max, dim, keepdims=True) + axis_min
    return tf.cast(masked_max, dtype=tf.float32), arg_max


def masked_min(distances: FloatTensor,
               mask: BoolTensor,
               dim: int = 1) -> Tuple[FloatTensor, FloatTensor]:
    """Computes the minimal values over masked pairwise distances.

    Args:
      distances: 2-D float `Tensor` of [n, n] pairwise distances
      mask: 2-D Boolean `Tensor` of [n, n] valid distance size.
      dim: The dimension over which to compute the minimum. Defaults to 1.

    Returns:
      A Tuple of Tensors containing the minimal distance value and the arg_min
      for each example.
    """
    # Convert to dbl to avoid precision error in offset
    distances = tf.cast(distances, dtype=tf.float64)
    mask = tf.cast(mask, dtype=tf.float64)

    axis_max = tf.math.reduce_max(distances, dim, keepdims=True) + 1e-6
    masked_min = tf.math.multiply(distances - axis_max, mask)
    arg_min = tf.math.argmin(masked_min, dim)
    masked_min = tf.math.reduce_min(masked_min, dim, keepdims=True) + axis_max
    return tf.cast(masked_min, dtype=tf.float32), arg_min


def build_masks(labels: IntTensor,
                batch_size: int) -> Tuple[BoolTensor, BoolTensor]:
    """Build masks that allows to select only the positive or negatives
    embeddings.
    Args:
        labels: 1D int `Tensor` that contains the class ids.
        batch_size: size of the batch.

    Returns:
        Tuple of Tensors containing the positive_mask and negative_mask
    """
    if tf.rank(labels) == 1:
        labels = tf.reshape(labels, (-1, 1))

    # same class mask
    positive_mask = tf.math.equal(labels, tf.transpose(labels))
    # not the same class
    negative_mask = tf.math.logical_not(positive_mask)

    # we need to remove the diagonal from positive mask
    diag = tf.logical_not(tf.linalg.diag(tf.ones(batch_size, dtype=tf.bool)))
    positive_mask = tf.math.logical_and(positive_mask, diag)

    return positive_mask, negative_mask



import tensorflow as tf

from tensorflow_similarity.models import SimilarityModel
from tensorflow_similarity.callbacks import EvalCallback
from tensorflow_similarity.callbacks import SplitValidationLoss


def test_eval_callback(tmp_path):
    queries = tf.constant([[1, 2], [1, 2]])
    query_labels = tf.constant([1, 2])
    targets = tf.constant([[1, 2], [1, 2]])
    target_labels = tf.constant([1, 2])

    log_dir = tmp_path / 'sec/'
    if not log_dir.exists():
        log_dir.mkdir(parents=True)

    callback = EvalCallback(
            queries=queries,
            query_labels=query_labels,
            targets=targets,
            target_labels=target_labels,
            tb_logdir=str(log_dir)
    )

    # manually set model ^^
    tf.keras.backend.clear_session()
    inputs = tf.keras.layers.Input(shape=(2, ))
    outputs = tf.keras.layers.Dense(2)(inputs)
    model = SimilarityModel(inputs, outputs)
    model.compile('adam', loss='mse', distance='cosine')
    callback.model = model

    # call the only callback method implemented
    callback.on_epoch_end(0, {})


def test_split_val_loss_callback(tmp_path):
    queries = tf.constant([[1, 2], [1, 2]])
    query_labels = tf.constant([1, 2])
    targets = tf.constant([[1, 2], [1, 2]])
    target_labels = tf.constant([1, 2])
    known_classes = tf.constant([1])

    log_dir = tmp_path / 'sec/'
    if not log_dir.exists():
        log_dir.mkdir(parents=True)

    callback = SplitValidationLoss(
            queries=queries,
            query_labels=query_labels,
            targets=targets,
            target_labels=target_labels,
            known_classes=known_classes,
            tb_logdir=str(log_dir)
    )

    # manually set model ^^
    tf.keras.backend.clear_session()
    inputs = tf.keras.layers.Input(shape=(2, ))
    outputs = tf.keras.layers.Dense(2)(inputs)
    model = SimilarityModel(inputs, outputs)
    model.compile('adam', loss='mse', distance='cosine')
    callback.model = model

    # call the only callback method implemented
    callback.on_epoch_end(0, {})



import numpy as np
import tensorflow as tf

from tensorflow_similarity.matchers import MatchNearest


def test_predict():
    mn = MatchNearest()

    lookup_labels = tf.constant([[10, 12], [20, 13], [30, 14], [40, 15]])
    lookup_distances = tf.constant(
        [[1.0, 1.5], [1.0, 1.7], [2.0, 2.1], [2.0, 2.2]]
    )

    d_labels, d_dist = mn.derive_match(lookup_labels, lookup_distances)

    np.testing.assert_array_equal(
        d_labels.numpy(), np.array([[10], [20], [30], [40]])
    )
    np.testing.assert_array_equal(
        d_dist.numpy(), np.array([[1.0], [1.0], [2.0], [2.0]])
    )



import numpy as np
import pytest
import tensorflow as tf

from tensorflow_similarity.retrieval_metrics import PrecisionAtK

testdata = [
    (
        "micro",
        tf.constant(0.583333333),
    ),
    (
        "macro",
        tf.constant(0.5),
    ),
]


@pytest.mark.parametrize("avg, expected", testdata, ids=["micro", "macro"])
def test_compute(avg, expected):
    query_labels = tf.constant([1, 1, 1, 0])
    match_mask = tf.constant(
        [
            [True, True, False],
            [True, True, False],
            [True, True, False],
            [False, False, True],
        ],
        dtype=bool,
    )
    rm = PrecisionAtK(k=3, average=avg)

    precision = rm.compute(query_labels=query_labels, match_mask=match_mask)
    np.testing.assert_allclose(precision, expected)



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf
from tensorflow_similarity.algebra import masked_max, masked_min
from tensorflow_similarity.types import FloatTensor, BoolTensor
from typing import Any, Tuple


@tf.keras.utils.register_keras_serializable(package="Similarity")
@tf.function
def positive_distances(positive_mining_strategy: str,
                       distances: FloatTensor,
                       positive_mask: BoolTensor
                       ) -> Tuple[FloatTensor, FloatTensor]:
    """Positive distance computation.

    Args:
        positive_mining_strategy (str, optional): [description].
        distances: 2-D float `Tensor` of [n, n] pairwise distances.
        positive_mask: 2-D Boolean `Tensor` of [n, n] valid distance size.

    Raises:
        ValueError: Invalid positive mining strategy

    Returns:
      A Tuple of Tensors containing the positive distance values
      and the index for each example.
    """
    if positive_mining_strategy == 'hard':
        positive_distances, pos_idxs = (
                masked_max(distances, positive_mask))
    elif positive_mining_strategy == 'easy':
        positive_distances, pos_idxs = (
                masked_min(distances, positive_mask))
    else:
        raise ValueError('Invalid positive mining strategy')

    return positive_distances, pos_idxs


@tf.keras.utils.register_keras_serializable(package="Similarity")
@tf.function
def negative_distances(negative_mining_strategy: str,
                       distances: FloatTensor,
                       negative_mask: BoolTensor,
                       positive_mask: BoolTensor,
                       batch_size: int) -> Tuple[FloatTensor, FloatTensor]:
    """Negative distance computation.

    Args:
        negative_mining_strategy: What mining strategy to use for select the
        embedding from the different class.
        Available: {'hard', 'semi-hard', 'easy'}

        distances: 2-D float `Tensor` of [n, n] pairwise distances.

        negative_mask: 2-D Boolean `Tensor` of [n, n] valid distance size.

        positive_mask: 2-D Boolean `Tensor` of [n, n] valid distance size.

        batch_size: The current batch size.

    Raises:
        ValueError: Invalid negative mining strategy

    Returns:
      A Tuple of Tensors containing the negative distance values
      and the index for each example.
    """
    if negative_mining_strategy == 'hard':
        # find the *non-zero* minimal distance between negative labels
        negative_distances, neg_idxs = (
                masked_min(distances, negative_mask))
    elif negative_mining_strategy == 'semi-hard':
        # find the minimal distance between negative label gt than max distance
        # between positive labels
        # find max value of positive distance
        max_positive, _ = masked_max(distances, positive_mask)

        # select distance that are above the max positive distance
        greater_distances = tf.math.greater(distances, max_positive)

        # combine with negative mask: keep negative value if greater,
        # zero otherwise
        empty = tf.zeros((batch_size, batch_size), dtype=tf.bool)
        semi_hard_mask = tf.where(greater_distances, negative_mask, empty)

        # find the  minimal distance between negative labels above threshold
        negative_distances, neg_idxs = (
                masked_min(distances, semi_hard_mask))

    elif negative_mining_strategy == 'easy':
        # find the maximal distance between negative labels
        negative_distances, neg_idxs = (
                masked_max(distances, negative_mask))
    else:
        raise ValueError('Invalid negative mining strategy')

    return negative_distances, neg_idxs


@tf.keras.utils.register_keras_serializable(package="Similarity")
@tf.function
def compute_loss(positive_distances: FloatTensor,
                 negative_distances: FloatTensor,
                 soft_margin: bool,
                 margin: float) -> Any:
    """Compute the final loss.

    Args:
        positive_distances: An [n,1] FloatTensor of positive distances.

        negative_distances: An [n,1] FloatTensor of negative distances.

        soft_margin: [description]. Defaults to False.

        margin: [description]. Defaults to 1.0.

    Returns:
        An [n,1] FloatTensor containing the loss for each example.
    """

    loss = tf.math.subtract(positive_distances, negative_distances)

    if soft_margin:
        loss = tf.reduce_logsumexp(loss)
    else:
        loss = tf.math.add(loss, margin)
        loss = tf.maximum(loss, 0.0)  # numeric stability

    return loss


@tf.keras.utils.register_keras_serializable(package="Similarity")
@tf.function
def logsumexp(pairwise_distances: FloatTensor,
              mask: FloatTensor) -> Any:
    """Compute the LogSumExp across axis 1 of the pairwise distance matrix.

    This function:
    * Avoids numerical instablity when the inputs are large or small.
    * Adds 1 to the reduce_sum of the exp to ensure the loss is positive
    * masks out the result of exp to ensure that masked values are not included
      in the log sum.

    Args:
        pairwise_distance: A 2D FloatTensor of the pairwise distances.

        mask: A 2D FloatTensor with 1.0 for all valid values and 0.0
        everywhere else.

    returns:
        A [n, 1] FloatTensor containing the per example LogSumExp values.
    """
    raw_max = tf.math.reduce_max(
        pairwise_distances,
        axis=1,
        keepdims=True
    )
    my_max = tf.stop_gradient(
        tf.where(
            tf.math.is_finite(raw_max),
            raw_max,
            tf.zeros_like(raw_max)
        )
    )
    x = tf.math.subtract(pairwise_distances, my_max)
    x = tf.math.exp(x) * mask
    x = tf.math.reduce_sum(x, axis=1, keepdims=True)
    x = tf.math.log(1 + x)
    x = tf.math.add(x, my_max)

    return x



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Index the embeddings infered by the model to allow distance based
sub-linear search"""

from collections import defaultdict, deque
import json
from pathlib import Path
from time import time
from typing import DefaultDict, Deque, Dict, List, Mapping, MutableMapping
from typing import Optional, Sequence, Union
import numpy as np
from tabulate import tabulate
import tensorflow as tf
from tqdm.auto import tqdm

# internal
from .distances import distance_canonicalizer, Distance
from .classification_metrics import ClassificationMetric
from .classification_metrics import F1Score
from .classification_metrics import make_classification_metric
from .evaluators import Evaluator, MemoryEvaluator
from .matchers import ClassificationMatch, make_classification_matcher
from .retrieval_metrics import RetrievalMetric
from .search import Search, NMSLibSearch
from .stores import Store, MemoryStore
from .utils import unpack_lookup_distances, unpack_lookup_labels
from .types import IntTensor, FloatTensor, Lookup, PandasDataFrame
from .types import Tensor, CalibrationResults


class Indexer():
    """Indexing system that allows to efficiently find nearest embeddings
    by indexing known embeddings and make them searchable using an
    [Approximate Nearest Neighbors Search]
    (https://en.wikipedia.org/wiki/Nearest_neighbor_search)
    search implemented via the [`Search()`](search/overview.md) classes
    and associated data lookup via the [`Store()`](stores/overview.md) classes.

    The indexer allows to evaluate the quality of the constructed index and
    calibrate the [SimilarityModel.match()](similarity_model.md) function via
    the [`Evaluator()`](evaluators/overview.md) classes.
    """
    # semantic sugar for the order of the returned data
    EMBEDDINGS = 0
    DISTANCES = 1
    LABELS = 2
    DATA = 3
    RANKS = 4

    def __init__(self,
                 embedding_size: int,
                 distance: Union[Distance, str] = 'cosine',
                 search: Union[Search, str] = 'nmslib',
                 kv_store: Union[Store, str] = 'memory',
                 evaluator: Union[Evaluator, str] = 'memory',
                 embedding_output: int = None,
                 stat_buffer_size: int = 1000) -> None:
        """Index embeddings to make them searchable via KNN

        Args:
            embedding_size: Size of the embeddings that will be stored.
            It is usually equivalent to the size of the output layer.

            distance: Distance used to compute embeddings proximity.
            Defaults to 'cosine'.

            kv_store: How to store the indexed records.
            Defaults to 'memory'.

            search: Which `Search()` framework to use to perform KNN
            search. Defaults to 'nmslib'.

            evaluator: What type of `Evaluator()` to use to evaluate index
            performance. Defaults to in-memory one.

            embedding_output: Which model output head predicts the embeddings
            that should be indexed. Default to None which is for single output
            model. For multi-head model, the callee, usually the
            `SimilarityModel()` class is responsible for passing the correct
            one.

            stat_buffer_size: Size of the sliding windows
            buffer used to compute index performance. Defaults to 1000.

        Raises:
            ValueError: Invalid search framework or key value store.
        """
        distance = distance_canonicalizer(distance)
        self.distance = distance  # needed for save()/load()
        self.embedding_output = embedding_output
        self.embedding_size = embedding_size

        # internal structure naming
        # FIXME support custom objects
        self.search_type = search
        self.kv_store_type = kv_store
        self.evaluator_type = evaluator

        # stats configuration
        self.stat_buffer_size = stat_buffer_size

        # calibration
        self.is_calibrated = False
        self.calibration_metric: ClassificationMetric = F1Score()
        self.cutpoints: Mapping[str, Mapping[str, Union[int, float, str]]] = {}
        self.calibration_thresholds: Mapping[str, Sequence[Union[float, int]]] = {}  # noqa

        # initialize internal structures
        self._init_structures()

    def reset(self) -> None:
        "Reinitialize the indexer"
        self._init_structures()

    def _init_structures(self) -> None:
        "(re)initialize internal storage structure"

        if self.search_type == 'nmslib':
            self.search: Search = NMSLibSearch(distance=self.distance,
                                               dims=self.embedding_size)
        elif isinstance(self.search_type, Search):
            self.search = self.search_type
        else:
            raise ValueError("You need to either supply a known search "
                             "framework name or a Search() object")

        # mapper from id to record data
        if self.kv_store_type == 'memory':
            self.kv_store: Store = MemoryStore()
        elif isinstance(self.kv_store_type, Store):
            self.kv_store = self.kv_store_type
        else:
            raise ValueError("You need to either supply a know key value "
                             "store name or a Store() object")

        # code used to evaluate indexer performance
        if self.evaluator_type == 'memory':
            self.evaluator: Evaluator = MemoryEvaluator()
        elif isinstance(self.evaluator_type, Evaluator):
            self.evaluator = self.evaluator_type
        else:
            raise ValueError("You need to either supply a know evaluator name "
                             "or an Evaluator() object")
        # stats
        self._stats: DefaultDict[str, int] = defaultdict(int)
        self._lookup_timings_buffer: Deque = deque([], maxlen=self.stat_buffer_size)  # noqa

        # calibration data
        self.is_calibrated = False
        self.calibration_metric = F1Score()
        self.cutpoints = {}
        self.calibration_thresholds = {}

    def _get_embedding(self, prediction: FloatTensor) -> FloatTensor:
        """Return the 1st embedding vector from a (multi-output) model
        prediction

        See: `single_lookup()`, `add()`

        Args:
            prediction: TF similarity model prediction, may be a multi-headed
            output.

        Returns:
            FloatTensor: 1D Tensor that contains the actual embedding
        """

        if isinstance(self.embedding_output, int):
            # in multi-output: embedding is [output_num][0]
            embedding: FloatTensor = prediction[self.embedding_output][0]
        else:
            # single output > return 1st element
            embedding = prediction[0]
        return embedding

    def _get_embeddings(self, predictions: FloatTensor) -> FloatTensor:
        """Return the embedding vectors from a (multi-output) model prediction

        Args:
            predictions: TF similarity model predictions, may be a multi-headed
            output.

        Returns:
            Tensor: 2D Tensor (num_embeddings, embedding_value)
        """

        if isinstance(self.embedding_output, int):
            embeddings: FloatTensor = predictions[self.embedding_output]
        else:
            # needed for typing
            embeddings = predictions
        return embeddings

    def _cast_label(self, label: Optional[int]) -> Optional[int]:
        if label is not None:
            label = int(label)
        return label

    def add(self,
            prediction: FloatTensor,
            label: Optional[int] = None,
            data: Tensor = None,
            build: bool = True,
            verbose: int = 1):
        """ Add a single embedding to the indexer

        Args:
            prediction: TF similarity model prediction, may be a multi-headed
            output.

            label: Label(s) associated with the
            embedding. Defaults to None.

            data: Input data associated with
            the embedding. Defaults to None.

            build: Rebuild the index after insertion.
            Defaults to True. Set it to false if you would like to add
            multiples batches/points and build it manually once after.

            verbose: Display progress if set to 1.
            Defaults to 1.
        """

        # deal with potential multi-output
        embedding = self._get_embedding(prediction)

        # store data and get its id
        idx = self.kv_store.add(embedding, label, data)

        # add index to the embedding
        self.search.add(embedding, idx, build=build, verbose=verbose)

    def batch_add(self,
                  predictions: FloatTensor,
                  labels: Optional[Sequence[int]] = None,
                  data: Optional[Tensor] = None,
                  build: bool = True,
                  verbose: int = 1):
        """Add a batch of embeddings to the indexer

        Args:
            predictions: TF similarity model predictions, may be a multi-headed
            output.

            labels: label(s) associated with the embedding. Defaults to None.

            datas: input data associated with the embedding. Defaults to None.

            build: Rebuild the index after insertion.
            Defaults to True. Set it to false if you would like to add
            multiples batches/points and build it manually once after.

            verbose: Display progress if set to 1. Defaults to 1.
        """

        # deal with potential multi-output
        embeddings = self._get_embeddings(predictions)

        # store points
        if verbose:
            print('|-Storing data points in key value store')
        idxs = self.kv_store.batch_add(embeddings, labels, data)
        self.search.batch_add(embeddings, idxs, build=build, verbose=verbose)

    def single_lookup(
            self,
            prediction: FloatTensor,
            k: int = 5) -> List[Lookup]:
        """Find the k closest matches of a given embedding

        Args:
            prediction: TF similarity model prediction, may be a multi-headed
            output.

            k: Number of nearest neighbors to lookup. Defaults to 5.
        Returns
            list of the k nearest neighbors info:
            List[Lookup]
        """

        embedding = self._get_embedding(prediction)
        start = time()
        idxs, distances = self.search.lookup(embedding, k=k)
        nn_embeddings, labels, data = self.kv_store.batch_get(idxs)

        lookup_time = time() - start
        lookups = []
        for i in range(len(nn_embeddings)):
            # ! casting is needed to avoid slowness down the line
            lookups.append(Lookup(
                rank=i + 1,
                embedding=nn_embeddings[i],
                distance=float(distances[i]),
                label=self._cast_label(labels[i]),
                data=data[i]
                ))
        self._lookup_timings_buffer.append(lookup_time)
        self._stats['num_lookups'] += 1
        return lookups

    def batch_lookup(self,
                     predictions: FloatTensor,
                     k: int = 5,
                     verbose: int = 1) -> List[List[Lookup]]:

        """Find the k closest matches for a set of embeddings

        Args:
            predictions: TF similarity model predictions, may be a multi-headed
            output.

            k: Number of nearest neighbors to lookup. Defaults to 5.

            verbose: Be verbose. Defaults to 1.

        Returns
            list of list of k nearest neighbors:
            List[List[Lookup]]
        """

        embeddings = self._get_embeddings(predictions)
        num_embeddings = len(embeddings)
        start = time()
        batch_lookups = []

        if verbose:
            print("\nPerforming NN search\n")
        batch_idxs, batch_distances = (
                self.search.batch_lookup(embeddings, k=k))

        if verbose:
            pb = tqdm(total=num_embeddings, desc='Building NN list')
        for eidx in range(num_embeddings):
            lidxs = batch_idxs[eidx]   # list of nn idxs
            distances = batch_distances[eidx]

            nn_embeddings, labels, data = self.kv_store.batch_get(lidxs)
            lookups = []
            for i in range(len(nn_embeddings)):
                # ! casting is needed to avoid slowness down the line
                lookups.append(Lookup(
                    rank=i + 1,
                    embedding=nn_embeddings[i],
                    distance=float(distances[i]),
                    label=self._cast_label(labels[i]),
                    data=data[i]
                ))
            batch_lookups.append(lookups)

            if verbose:
                pb.update()

        if verbose:
            pb.close()

        # stats
        lookup_time = time() - start
        per_lookup_time = lookup_time / num_embeddings
        for _ in range(num_embeddings):
            self._lookup_timings_buffer.append(per_lookup_time)
        self._stats['num_lookups'] += num_embeddings

        return batch_lookups

    # evaluation related functions
    def evaluate_retrieval(
            self,
            predictions: FloatTensor,
            target_labels: Sequence[int],
            retrieval_metrics: Sequence[RetrievalMetric],
            verbose: int = 1) -> Dict[str, np.ndarray]:
        """Evaluate the quality of the index against a test dataset.

        Args:
            predictions: TF similarity model predictions, may be a multi-headed
            output.

            target_labels: Sequence of the expected labels associated with the
            embedded queries.

            retrieval_metrics: List of
            [RetrievalMetric()](retrieval_metrics/overview.md) to compute.

            verbose (int, optional): Display results if set to 1 otherwise
            results are returned silently. Defaults to 1.

        Returns:
            Dictionary of metric results where keys are the metric names and
            values are the metrics values.
        """
        # Determine the maximum number of neighbors needed by the retrieval
        # metrics because we do a single lookup.
        k = 1
        for m in retrieval_metrics:
            if not isinstance(m, RetrievalMetric):
                raise ValueError(m, 'is not a valid RetrivalMetric(). The '
                                 'RetrivialMetric() must be instantiated with '
                                 'a valid K.')
            if m.k > k:
                k = m.k

        # Find NN
        lookups = self.batch_lookup(predictions, k=k, verbose=verbose)

        # Evaluate them
        return self.evaluator.evaluate_retrieval(
                retrieval_metrics=retrieval_metrics,
                target_labels=target_labels,
                lookups=lookups)

    def evaluate_classification(
            self,
            predictions: FloatTensor,
            target_labels: Union[Sequence[int], IntTensor],
            distance_thresholds: Union[Sequence[float], FloatTensor],
            metrics: Sequence[Union[str, ClassificationMetric]] = ['f1'],
            matcher: Union[str, ClassificationMatch] = 'match_nearest',
            k: int = 1,
            verbose: int = 1) -> Dict[str, np.ndarray]:
        """Evaluate the classification performance.

        Compute the classification metrics given a set of queries, lookups, and
        distance thresholds.

        Args:
            predictions: TF similarity model predictions, may be a multi-headed
            output.

            target_labels: Sequence of expected labels for the lookups.

            distance_thresholds: A 1D tensor denoting the distances points at
            which we compute the metrics.

            metrics: The set of classification metrics.

            matcher: {'match_nearest', 'match_majority_vote'} or
            ClassificationMatch object. Defines the classification matching,
            e.g., match_nearest will count a True Positive if the query_label
            is equal to the label of the nearest neighbor and the distance is
            less than or equal to the distance threshold.

            distance_rounding: How many digit to consider to
            decide if the distance changed. Defaults to 8.

            verbose: Be verbose. Defaults to 1.
        Returns:
            A Mapping from metric name to the list of values computed for each
            distance threshold.
        """
        combined_metrics: List[ClassificationMetric] = (
            [make_classification_metric(m) for m in metrics])

        lookups = self.batch_lookup(predictions, k=k, verbose=verbose)

        lookup_distances = unpack_lookup_distances(lookups)
        lookup_labels = unpack_lookup_labels(lookups)
        query_labels: IntTensor = tf.cast(
                tf.convert_to_tensor(target_labels),
                dtype='int32'
        )
        thresholds: FloatTensor = tf.cast(
                tf.convert_to_tensor(distance_thresholds),
                dtype='float32'
        )

        results = self.evaluator.evaluate_classification(
            query_labels=query_labels,
            lookup_labels=lookup_labels,
            lookup_distances=lookup_distances,
            distance_thresholds=thresholds,
            metrics=combined_metrics,
            matcher=matcher,
            verbose=verbose)

        return results

    def calibrate(self,
                  predictions: FloatTensor,
                  target_labels: Sequence[int],
                  thresholds_targets: MutableMapping[str, float],
                  calibration_metric: Union[str, ClassificationMetric] = "f1_score",  # noqa
                  k: int = 1,
                  matcher: Union[str, ClassificationMatch] = 'match_nearest',
                  extra_metrics: Sequence[Union[str, ClassificationMetric]] = ['precision', 'recall'],  # noqa
                  rounding: int = 2,
                  verbose: int = 1) -> CalibrationResults:
        """Calibrate model thresholds using a test dataset.

        FIXME: more detailed explanation.

        Args:
            predictions: TF similarity model predictions, may be a multi-headed
            output.

            target_labels: Sequence of the expected labels associated with the
            embedded queries.

            thresholds_targets: Dict of performance targets to (if possible)
            meet with respect to the `calibration_metric`.

            calibration_metric: [ClassificationMetric()](metrics/overview.md)
            used to evaluate the performance of the index.

            k: How many neighbors to use during the calibration.
            Defaults to 1.

            matcher: {'match_nearest', 'match_majority_vote'} or
            ClassificationMatch object. Defines the classification matching,
            e.g., match_nearest will count a True Positive if the query_label
            is equal to the label of the nearest neighbor and the distance is
            less than or equal to the distance threshold.
            Defaults to 'match_nearest'.

            extra_metrics: List of additional
            `tf.similarity.classification_metrics.ClassificationMetric()` to
            compute and report. Defaults to ['precision', 'recall'].

            rounding: Metric rounding. Default to 2 digits.

            verbose: Be verbose and display calibration results. Defaults to 1.

        Returns:
            CalibrationResults containing the thresholds and cutpoints Dicts.
        """

        # find NN
        lookups = self.batch_lookup(predictions, k=k, verbose=verbose)

        # making sure our metrics are all ClassificationMetric objects
        calibration_metric = make_classification_metric(calibration_metric)

        combined_metrics: List[ClassificationMetric] = (
            [make_classification_metric(m) for m in extra_metrics])

        # running calibration
        calibration_results = self.evaluator.calibrate(
            target_labels=target_labels,
            lookups=lookups,
            thresholds_targets=thresholds_targets,
            calibration_metric=calibration_metric,
            matcher=matcher,
            extra_metrics=combined_metrics,
            metric_rounding=rounding,
            verbose=verbose
        )

        # display cutpoint results if requested
        if verbose:
            headers = ['name', 'value', 'distance']  # noqa
            cutpoints = list(calibration_results.cutpoints.values())
            # dynamically find which metrics we need. We only need to look at
            # the first cutpoints dictionary as all subsequent ones will have
            # the same metric keys.
            for metric_name in cutpoints[0].keys():
                if metric_name not in headers:
                    headers.append(metric_name)

            rows = []
            for data in cutpoints:
                rows.append([data[v] for v in headers])
            print("\n", tabulate(rows, headers=headers))

        # store info for serialization purpose
        self.is_calibrated = True
        self.calibration_metric = calibration_metric
        self.cutpoints = calibration_results.cutpoints
        self.calibration_thresholds = calibration_results.thresholds
        return calibration_results

    def match(self,
              predictions: FloatTensor,
              no_match_label: int = -1,
              k=1,
              matcher: Union[str, ClassificationMatch] = 'match_nearest',
              verbose: int = 1) -> Dict[str, List[int]]:
        """Match embeddings against the various cutpoints thresholds

        Args:
            predictions: TF similarity model predictions, may be a multi-headed
            output.

            no_match_label: What label value to assign when there is no match.
            Defaults to -1.

            k: How many neighboors to use during the calibration.
            Defaults to 1.

            matcher: {'match_nearest', 'match_majority_vote'} or
            ClassificationMatch object. Defines the classification matching,
            e.g., match_nearest will count a True Positive if the query_label
            is equal to the label of the nearest neighbor and the distance is
            less than or equal to the distance threshold.

            verbose: display progression. Default to 1.

        Notes:

            1. It is up to the [`SimilarityModel.match()`](similarity_model.md)
            code to decide which of cutpoints results to use / show to the
            users. This function returns all of them as there is little
            performance downside to do so and it makes the code clearer
            and simpler.

            2. The calling function is responsible to return the list of class
            matched to allows implementation to use additional criteria if they
            choose to.

        Returns:
            Dict of cutpoint names mapped to lists of matches.
        """
        matcher = make_classification_matcher(matcher)

        lookups = self.batch_lookup(predictions, k=k, verbose=verbose)

        lookup_distances = unpack_lookup_distances(lookups)
        lookup_labels = unpack_lookup_labels(lookups)

        if verbose:
            pb = tqdm(total=len(lookup_distances) * len(self.cutpoints),
                      desc='matching embeddings')

        matches: DefaultDict[str, List[int]] = defaultdict(list)
        for cp_name, cp_data in self.cutpoints.items():
            distance_threshold = float(cp_data['distance'])

            pred_labels, pred_dist = matcher.derive_match(
                    lookup_labels=lookup_labels,
                    lookup_distances=lookup_distances)

            for label, distance in zip(pred_labels, pred_dist):
                if distance <= distance_threshold:
                    label = int(label)
                else:
                    label = no_match_label

                matches[cp_name].append(label)

                if verbose:
                    pb.update()

        if verbose:
            pb.close()

        return matches

    def save(self, path: str, compression: bool = True):
        """Save the index to disk

        Args:
            path: directory where to save the index
            compression: Store index data compressed. Defaults to True.
        """
        path = str(path)
        # saving metadata
        metadata = {
            "size": self.size(),
            "compression": compression,
            "distance": self.distance.name,
            "embedding_output": self.embedding_output,
            "embedding_size": self.embedding_size,

            "kv_store": self.kv_store_type,
            "evaluator": self.evaluator_type,
            "search": self.search_type,

            "stat_buffer_size": self.stat_buffer_size,
            "is_calibrated": self.is_calibrated,

            "calibration_metric_config": self.calibration_metric.get_config(),
            "cutpoints": self.cutpoints,
            "calibration_thresholds": self.calibration_thresholds
        }

        metadata_fname = self.__make_metadata_fname(path)
        tf.io.write_file(metadata_fname, json.dumps(metadata))

        self.kv_store.save(path, compression=compression)
        self.search.save(path)

    @staticmethod
    def load(path: Union[str, Path], verbose: int = 1):
        """Load Index data from a checkpoint and initialize underlying
        structure with the reloaded data.

        Args:
            path: Directory where the checkpoint is located.
            verbose: Be verbose. Defaults to 1.

        Returns:
            Initialized index
        """
        path = str(path)
        # recreate the index from metadata
        metadata_fname = Indexer.__make_metadata_fname(path)
        metadata = tf.io.read_file(metadata_fname)
        metadata = tf.keras.backend.eval(metadata)
        md = json.loads(metadata)
        index = Indexer(distance=md['distance'],
                        embedding_size=md['embedding_size'],
                        embedding_output=md['embedding_output'],
                        kv_store=md['kv_store'],
                        evaluator=md['evaluator'],
                        search=md['search'],
                        stat_buffer_size=md['stat_buffer_size'])

        # reload the key value store
        if verbose:
            print("Loading index data")
        index.kv_store.load(path)

        # rebuild the index
        if verbose:
            print('Loading search index')
        index.search.load(path)

        # reload calibration data if any
        index.is_calibrated = md['is_calibrated']
        if index.is_calibrated:
            if verbose:
                print("Loading calibration data")
            index.calibration_metric = make_classification_metric(
                    metric=md['calibration_metric_config']['canonical_name'],
                    name=md['calibration_metric_config']['name'])

            index.cutpoints = md['cutpoints']
            index.calibration_thresholds = md['calibration_thresholds']
        return index

    def get_calibration_metric(self):
        return self.calibration_metric

    def size(self) -> int:
        "Return the index size"
        return self.kv_store.size()

    def stats(self):
        """return index statistics"""
        stats = self._stats
        stats['size'] = self.kv_store.size()
        stats['stat_buffer_size'] = self.stat_buffer_size

        # query performance - make sure we don't count unused buffer
        max_idx = min(stats['num_lookups'], self.stat_buffer_size)
        lookup_timings = list(self._lookup_timings_buffer)[:max_idx]

        # ensure we never have an empty list
        lookup_timings = lookup_timings if list(lookup_timings) else [0]

        # compute stats
        stats['query_performance'] = {
            'min': np.min(lookup_timings),
            'max': np.max(lookup_timings),
            'avg': np.average(lookup_timings),
            'median': np.median(lookup_timings),
            'stddev': np.std(lookup_timings)
        }
        return stats

    def print_stats(self):
        "display statistics in terminal friendly fashion"
        # compute statistics
        stats = self.stats()

        # info
        print('[Info]')
        rows = [
            ['distance', self.distance],
            ['key value store', self.kv_store_type],
            ['search algorithm', self.search_type],
            ['evaluator', self.evaluator_type],
            ['index size', self.size()],
            ['calibrated', self.is_calibrated],
            ['calibration_metric', self.calibration_metric.name],
            ['embedding_output', self.embedding_output]
        ]
        print(tabulate(rows))
        print('\n')

        print('\n[Performance]')
        rows = [['num lookups', stats['num_lookups']]]
        for k, v in stats['query_performance'].items():
            rows.append([k, v])
        print(tabulate(rows))

    def to_data_frame(self, num_items: int = 0) -> PandasDataFrame:
        """Export data as pandas dataframe

        Args:
            num_items (int, optional): Num items to export to the dataframe.
            Defaults to 0 (unlimited).

        Returns:
            pd.DataFrame: a pandas dataframe.
        """
        return self.kv_store.to_data_frame(num_items)

    @staticmethod
    def __make_metadata_fname(path):
        return str(Path(path) / 'index_metadata.json')
